#!/usr/bin/env python3
"""
Plot form factors.
"""
import bornagain as ba
from bornagain import nm, deg
import bornplot as bp

det = ba.SphericalDetector(200, 5*deg, 2.5*deg, 2.5*deg)
n = 4
results = []
for i in range(n):
    omega = 30*i/(n - 1)
    title = r'$\omega=%d^\circ$' % omega
    ff = ba.FormFactorPrism3(13.8*nm, 3*nm)
    trafo = ba.RotationZ(omega*deg)
    data = bp.run_simulation(det, ff, trafo)
    results.append(bp.Result(i, data, title))

bp.make_plot(results, det, "ff_Prism3")



#!/usr/bin/env python3
"""
Long boxes at 1D lattice, ba.OffSpecular simulation
"""
import bornagain as ba
from bornagain import angstrom, deg, nm, nm2, kvector_t

phi_f_min, phi_f_max = -1, 1.0
alpha_f_min, alpha_f_max = 0, 10.0
alpha_i_min, alpha_i_max = 0, 10.0  # incoming beam


def get_sample():
    """
    Returns a sample with a grating on a substrate,
    modelled by infinitely long boxes forming a 1D lattice.
    """

    # Define materials
    material_Particle = ba.HomogeneousMaterial("Particle", 0.0006, 2e-08)
    material_Substrate = ba.HomogeneousMaterial("Substrate", 6e-06, 2e-08)
    material_Vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)

    # Define form factors
    ff = ba.FormFactorBox(1000*nm, 20*nm, 10*nm)

    # Define particles
    particle = ba.Particle(material_Particle, ff)
    particle_rotation = ba.RotationZ(90*deg)
    particle.setRotation(particle_rotation)

    # Define interference functions
    iff = ba.InterferenceFunction1DLattice(100*nm, 0)
    iff_pdf = ba.FTDecayFunction1DCauchy(1e6*nm)
    iff.setDecayFunction(iff_pdf)

    # Define particle layouts
    layout = ba.ParticleLayout()
    layout.addParticle(particle)
    layout.setInterferenceFunction(iff)
    layout.setTotalParticleSurfaceDensity(0.01)

    # Define layers
    layer_1 = ba.Layer(material_Vacuum)
    layer_1.addLayout(layout)
    layer_2 = ba.Layer(material_Substrate)

    # Define sample
    sample = ba.MultiLayer()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)

    return sample


def get_simulation(sample):
    """
    Returns an off-specular simulation with beam and detector defined.
    """
    simulation = ba.OffSpecularSimulation()
    simulation.setDetectorParameters(20, phi_f_min*deg, phi_f_max*deg, 200,
                                     alpha_f_min*deg, alpha_f_max*deg)
    # define the beam with alpha_i varied between alpha_i_min and alpha_i_max
    alpha_i_axis = ba.FixedBinAxis("alpha_i", 200, alpha_i_min*deg,
                                   alpha_i_max*deg)
    simulation.setBeamParameters(1*angstrom, alpha_i_axis, 0)
    simulation.beam().setIntensity(1e9)
    simulation.setSample(sample)
    return simulation


if __name__ == '__main__':
    import ba_plot
    sample = get_sample()
    simulation = get_simulation(sample)
    ba_plot.run_and_plot(simulation, intensity_min=1)



"""
Installs BornAgain libraries into user Python (Mac only).

Usage: python bornagain_python_install.py

The script generates BornAgain python package in temporary directory and then
installs it into user's Python site_packages.

Requirements: BornAgain.app has to be installed on the system using .dmg installer.

During generation of Python package
1) The script copies BornAgain libraries from the GUI installation directory (normally
   it is /Applications/BornAgain.app) into temporary bundle directory
2) Adjusts libraries using Mac's install_name_tool to rely on Python's own libpython2.7.dylib.
   The exact site-packages library location is deduced from the interpreter itself.
3) Generates other files necessary for Python packaging (setup.py, __init__.py)

During installation of Python package
1) The script just runs standard 'python setup.py install' command from the temporary bundle directory
"""

from __future__ import print_function
import os
import sys
import sysconfig
import glob
import tempfile
import shutil
import platform
import site
import subprocess
from distutils.sysconfig import get_python_lib

BORNAGAIN_VERSION = "0.0"


def is_python3():
    if (sys.version_info > (3, 0)):
        return True
    else:
        return False


def python_version_string():
    return str(sys.version_info[0]) + "." + str(sys.version_info[1])


def add_rpath(newpath, filename):
    print("add_rpath ", newpath, filename)
    p = subprocess.Popen(
        ['install_name_tool', '-add_rpath', newpath, filename],
        stdout=subprocess.PIPE)
    p.communicate()
    return


def change_rpath(oldpath, newpath, filename):
    print("change_rpath ", oldpath, newpath, filename)
    p = subprocess.Popen(
        ['install_name_tool', '-change', oldpath, newpath, filename],
        stdout=subprocess.PIPE)
    p.communicate()
    return


def delete_rpath(todelete, filename):
    print("delete_rpath ", todelete, filename)
    p = subprocess.Popen(
        ['install_name_tool', '-delete_rpath', todelete, filename],
        stdout=subprocess.PIPE)
    p.communicate()
    return


def exec_full(filepath):
    """
    Executes embedded python script.
    http://stackoverflow.com/questions/436198/what-is-an-alternative-to-execfile-in-python-3
    """
    import os
    global_namespace = {
        "__file__": filepath,
        "__name__": "__main__",
    }
    with open(filepath, 'rb') as file:
        exec(compile(file.read(), filepath, 'exec'), global_namespace)


def get_python_shared_library():
    """
    Returns full path to the python shared library on which current interpreter is relying
    """
    prefix = sys.prefix
    suffix = sysconfig.get_config_var(
        'LDVERSION') or sysconfig.get_config_var('VERSION')
    result = sys.prefix + "/lib/libpython" + suffix + ".dylib"
    return result


def get_script_path():
    """
    Returns the full path to the directory where given scripts resides
    """
    return os.path.dirname(os.path.realpath(sys.argv[0]))


def get_application_dir():
    """
    Returns BornAgain.app directory location, determines BornAgain version number
    """
    global BORNAGAIN_VERSION
    app_dir = ""
    if len(sys.argv) == 2:
        app_dir = sys.argv[1]
    else:
        script_dir = get_script_path()
        app_dir = os.path.abspath(
            os.path.join(script_dir, "..", "..", ".."))

    # getting version number
    lib_dir = glob.glob(
        os.path.join(app_dir, "Contents", "lib", "BornAgain-*"))
    if len(lib_dir) != 1:
        exit("Can't find BornAgain libraries in " + app_dir)

    BORNAGAIN_VERSION = lib_dir[0].split("BornAgain-")[1]
    return app_dir


def create_bundle_tem_dir():
    """
    Creates temporary directory for BornAgain package bundle
    """
    tmpdir = os.path.join(tempfile.gettempdir(), "bornagain_bundle")
    if os.path.exists(tmpdir):
        shutil.rmtree(tmpdir)
    os.makedirs(tmpdir)
    return tmpdir


def create_package_dir(destination_dir):
    """
    Create package directory in bundle directory
    """
    packagedir = os.path.join(destination_dir, "bornagain")
    os.makedirs(packagedir)
    return packagedir


def create_library_dir(destination_dir):
    """
    Create library directory in bundle directory
    """
    librarydir = os.path.join(destination_dir, "lib")
    os.makedirs(librarydir)
    return librarydir


def generate_setup_py(destination_dir):
    """
    Generates setup.py file in BornAgain's bundle directory
    """
    text = """\
# BornAgain setup.py to install BornAgain libraries into Python's site-packages
# Usage: python setup.py install

import os
import glob
from setuptools import setup


datadir = os.path.join('bornagain', 'lib','BornAgain-{0}')
datafiles = [(datadir, [f for f in glob.glob(os.path.join(datadir, '*'))])]

datadir = os.path.join('bornagain', 'lib','Frameworks')
datafiles += [(datadir, [f for f in glob.glob(os.path.join(datadir, '*'))])]

setup(name='bornagain',
    version='{0}',
    description='Simulate and fit grazing-incidence small-angle scattering',
    url='www.bornagainproject.org',
    author='Scientific Computing Group at MLZ',
    author_email='contact@bornagainproject.org',
    license='GPL',
    packages=['bornagain'],
    data_files = datafiles,
    zip_safe=False,
    )
""".format(BORNAGAIN_VERSION)
    with open(os.path.join(destination_dir, "setup.py"), "w") as fout:
        fout.write(text)
        fout.close()
    pass


def prepare_init_module(app_dir, bundle_dir):
    source_dir = os.path.join(app_dir, "Contents", "libexec")
    libexec_dir = os.path.join(source_dir,
                               "BornAgain-" + BORNAGAIN_VERSION,
                               "bornagain")
    package_dir = os.path.join(bundle_dir, "bornagain")
    print("--> Copying modules from '{0}' to '{1}'".format(
        libexec_dir, package_dir))
    shutil.copytree(libexec_dir, package_dir)
    return package_dir


def copy_libraries(app_dir, destination_dir):
    """
    Copy libraries from BornAgain.app into corresponding BornAgain Python package directory
    """
    print("--> Copying libraries from '{0}'".format(app_dir))
    app_bornagainlib_dir = os.path.join(app_dir, "Contents", "lib",
                                        "BornAgain-" + BORNAGAIN_VERSION)
    app_frameworks_dir = os.path.join(app_dir, "Contents", "Frameworks")

    # copying BornAgain libraries
    shutil.copytree(
        app_bornagainlib_dir,
        os.path.join(destination_dir, "BornAgain-" + BORNAGAIN_VERSION))

    # cleaning unnecessary files
    libfiles = glob.glob(
        os.path.join(destination_dir, '*/libBornAgainGUI*'))
    for f in libfiles:
        os.remove(f)
    pass

    # copying libraries from Frameworks
    frameworks_libs = glob.glob(os.path.join(app_frameworks_dir, 'lib*'))
    frameworks_dest = os.path.join(destination_dir, "Frameworks")
    if not os.path.exists(frameworks_dest):
        os.makedirs(frameworks_dest)
    for lib in frameworks_libs:
        shutil.copyfile(
            lib, os.path.join(frameworks_dest, os.path.basename(lib)))


def patch_libraries(dir_name):
    """
    Patches libraries depending on Python to point on the same shared libpython2.7.dylib which current interpreter is using
    """
    print("--> Patching libraries to rely on '{0}'".format(
        get_python_shared_library()))
    libfiles = glob.glob(os.path.join(dir_name, '*/_libBornAgain*'))
    for f in libfiles:
        for lib in ["Base", "Fit", "Param", "Sample", "Device", "Core"]:
            if "libBornAgain" + lib in f:
                delete_rpath("@loader_path/../../Frameworks", f)
        add_rpath("@loader_path/../Frameworks", f)
        if "libBornAgainCore" in f:
            add_rpath("@loader_path/.", f)

    libfiles += glob.glob(os.path.join(dir_name, '*/libboost_python*'))
    for f in libfiles:
        change_rpath(
            "@rpath/Python.framework/Versions/" + python_version_string() +
            "/Python", get_python_shared_library(), f)

    pass


def create_bundle(app_dir):
    """
    Creates ready to install BornAgain Python bundle package
    """
    bundle_dir = create_bundle_tem_dir()
    print('-'*80)
    print("Generating Python bundle in temporary '{0}'".format(bundle_dir))
    print('-'*80)

    print("--> Generating bundle setup files")

    generate_setup_py(bundle_dir)

    package_dir = prepare_init_module(app_dir, bundle_dir)

    library_dir = create_library_dir(package_dir)
    copy_libraries(app_dir, library_dir)

    patch_libraries(library_dir)

    print(
        "\nBornAgain Python bundle is successfully created in temporary directory '{0}'"
        .format(bundle_dir))
    print(
        "Run 'python setup.py install' from there to install it into your Python's site-packages"
    )
    return bundle_dir


def install_bundle(dir_name):
    """
    Installs BornAgain Python bundle previously generated in the directory dir_name
    """
    print('-'*80)
    print("Installing bundle in Python site-packages '{0}'".format(
        get_python_lib()))
    print('-'*80)

    os.chdir(bundle_dir)
    sys.argv = ['setup.py', 'install']
    exec_full('setup.py')
    print("\nBornAgain Python bundle is successfully installed in '{0}'".
          format(get_python_lib()))
    print("Congratulations!")


if __name__ == '__main__':
    if not platform.system() == 'Darwin':
        exit("This script is intended for MacOs systems. Exiting...")

    app_dir = get_application_dir()

    print('-'*80)
    print(
        "Installation of BornAgain-{0} libraries into site-packages of your Python"
        .format(BORNAGAIN_VERSION))
    print('-'*80)
    print("From :", app_dir)
    print("To   :", get_python_lib())
    print(" ")
    print("Possible options:")
    print(
        "[0] - Generate bundle with BornAgain libraries, do not install it."
    )
    print(
        "[1] - Generate bundle and install it into site-packages of your Python."
    )
    print("[2] - Exit")

    var = 0
    if is_python3():
        var = int(input("Enter your choice [1]: ") or "1")
    else:
        var = int(raw_input("Enter your choice [1]: ") or "1")

    if var == 0:
        create_bundle(app_dir)
    elif var == 1:
        bundle_dir = create_bundle(app_dir)
        install_bundle(bundle_dir)
    else:
        exit("Good bye")



#!/usr/bin/env python3
"""
Simulation of grating using very long boxes and 1D lattice.
Monte-carlo integration is used to get rid of
large-particle form factor oscillations.
"""
import bornagain as ba
from bornagain import deg, angstrom, nm, micrometer
import ba_plot
from matplotlib import pyplot as plt


def get_sample(lattice_rotation_angle=0*deg):
    """
    Returns a sample with a grating on a substrate.
    lattice_rotation_angle = 0 - beam parallel to grating lines
    lattice_rotation_angle = 90*deg - beam perpendicular to grating lines
    """
    # defining materials
    m_vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)
    m_si = ba.HomogeneousMaterial("Si", 5.7816e-6, 1.0229e-7)

    box_length, box_width, box_height = 50*micrometer, 70*nm, 50*nm
    lattice_length = 150*nm

    # collection of particles
    interference = ba.InterferenceFunction1DLattice(
        lattice_length, 90*deg - lattice_rotation_angle)

    pdf = ba.FTDecayFunction1DGauss(450)
    interference.setDecayFunction(pdf)

    box_ff = ba.FormFactorLongBoxLorentz(box_length, box_width, box_height)
    box = ba.Particle(m_si, box_ff)

    particle_layout = ba.ParticleLayout()
    particle_layout.addParticle(box, 1, ba.kvector_t(0, 0, 0),
                                ba.RotationZ(lattice_rotation_angle))
    particle_layout.setInterferenceFunction(interference)

    # assembling the sample
    vacuum_layer = ba.Layer(m_vacuum)
    vacuum_layer.addLayout(particle_layout)
    substrate_layer = ba.Layer(m_si)

    roughness = ba.LayerRoughness()
    roughness.setSigma(5*nm)
    roughness.setHurstParameter(0.5)
    roughness.setLatteralCorrLength(10*nm)

    multi_layer = ba.MultiLayer()
    multi_layer.addLayer(vacuum_layer)
    multi_layer.addLayerWithTopRoughness(substrate_layer, roughness)
    return multi_layer


def get_simulation(sample):
    """
    Create and return GISAXS simulation with beam and detector defined
    """
    beam = ba.Beam(1e8, 1.34*angstrom, ba.Direction(0.4*deg, 0))
    det = ba.SphericalDetector(200, -0.5*deg, 0.5*deg, 200, 0, 0.6*deg)
    simulation = ba.GISASSimulation(beam, sample, det)
    simulation.getOptions().setMonteCarloIntegration(True, 100)
    return simulation


def run_simulation():
    """
    Runs simulation and returns intensity map.
    """
    sample = get_sample()
    simulation = get_simulation(sample)
    if not "__no_terminal__" in globals():
        simulation.setTerminalProgressMonitor()
    simulation.runSimulation()
    return simulation.result()


def simulate_and_plot():
    interactive = True
    result = run_simulation().histogram2d()
    ba_plot.plot_histogram(result)

    peaks = ba.FindPeaks(result, 2, "nomarkov", 0.001)
    xpeaks = [peak[0] for peak in peaks]
    ypeaks = [peak[1] for peak in peaks]
    print(peaks)
    plt.plot(xpeaks,
             ypeaks,
             linestyle='None',
             marker='x',
             color='white',
             markersize=10)
    plt.show()


if __name__ == '__main__':
    simulate_and_plot()



"""
Utilities to plot form factors of particles in Born approximation
"""
import math, numpy
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib import rc
rc('font', **{'family': 'sans-serif', 'sans-serif': ['Helvetica']})
rc('text', usetex=True)
rc('image', cmap='inferno')

import bornagain as ba
from bornagain import deg, angstrom


class Result:
    def __init__(self, idx, data, title=""):
        self.idx = idx
        self.data = data
        self.title = title


def rectangle(det):
    return (det.axis(0).lowerBound(), det.axis(0).upperBound(),
            det.axis(1).lowerBound(), det.axis(1).upperBound())


def make_plot(results, det, name, nrow=1):
    """
    Make a plot consisting of one detector image for each Result in results,
    plus one common color scale.

    :param results: List of simulation results
    :param det: Detector
    :param name: Filename for plot during save
    :param nrow: Number of rows for different plots
    """
    mpl.rcParams['image.interpolation'] = 'none'
    n = len(results)
    ncol = 1 + (n - 1) // nrow
    # Parameters as fraction of subfig size.
    yskip = 0.2  # +ncol*0.02
    bottomskip = yskip
    topskip = yskip/2
    xskip = 0.18
    leftskip = xskip
    rightskip = 0.28 + ncol*0.03
    xtot = ncol*1.0 + (ncol - 1)*xskip + leftskip + rightskip
    ytot = nrow*1.0 + (nrow - 1)*yskip + bottomskip + topskip
    # We need parameters as fraction of total fig size.
    xskip /= xtot
    leftskip /= xtot
    rightskip /= xtot
    yskip /= ytot
    bottomskip /= ytot
    topskip /= ytot
    # Set total figure dimensions.
    ftot = 5
    fontsize = 18 + 36.0/(ncol + 2)
    # Create the figure 'fig' and its subplots axes ('tmp'->'axes').
    fig, tmp = plt.subplots(nrow, ncol, figsize=(ftot*xtot, ftot*ytot))
    if n > 1:
        axes = tmp.flat
    else:
        axes = [tmp]
    # Always the same color scale, to facilitate comparisons between figures.
    norm = mpl.colors.LogNorm(1e-8, 1)
    # Plot the subfigures.
    for res in results:
        ax = axes[res.idx]
        im = ax.imshow(res.data,
                       norm=norm,
                       extent=rectangle(det),
                       aspect=1)
        ax.set_xlabel(r'$\phi_{\rm f} (^{\circ})$', fontsize=fontsize)
        if res.idx % ncol == 0:
            ax.set_ylabel(r'$\alpha_{\rm f} (^{\circ})$',
                          fontsize=fontsize)
        if res.title != "":
            ax.set_title(res.title, fontsize=fontsize)
        ax.tick_params(axis='both',
                       which='major',
                       labelsize=fontsize*21/24)
    # Adjust whitespace around and between subfigures.
    plt.subplots_adjust(wspace=xskip,
                        hspace=yskip,
                        left=leftskip,
                        right=1 - rightskip,
                        bottom=bottomskip,
                        top=1 - topskip)
    # Plot the color scale.
    cbar_ax = fig.add_axes([
        1 - rightskip + 0.4*xskip, bottomskip, 0.25*xskip,
        1 - bottomskip - topskip
    ])
    cb = fig.colorbar(im, cax=cbar_ax)
    cb.set_label(r'$\left|F(q)\right|^2/V^{\,2}$', fontsize=fontsize)
    # Output to data file, image file, and display.
    nDigits = int(math.log10(len(results))) + 1
    formatN = "%" + str(nDigits) + "i"
    for i in range(len(results)):
        fname = name + "." + (formatN % i) + ".int"
        print(fname)
        numpy.savetxt(fname, results[i].data)
    plt.savefig(name + ".pdf", format="pdf", bbox_inches='tight')
    # plt.show()


def get_sample(ff, trafo):
    """Build and return a sample consisting of uncorrelated particles with given
    form factor and transformation operator

    :param ff: Form factor
    :param trafo: Optional rotation
    """
    # defining materials
    m_vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)
    m_particle = ba.HomogeneousMaterial("Particle", 1e-5, 0)

    # collection of particles
    particle = ba.Particle(m_particle, ff)
    particle_layout = ba.ParticleLayout()
    if trafo is not None:
        particle.setRotation(trafo)
        particle_layout.addParticle(particle)
    else:
        particle_layout.addParticle(particle)

    vacuum_layer = ba.Layer(m_vacuum)
    vacuum_layer.addLayout(particle_layout)

    multi_layer = ba.MultiLayer()
    multi_layer.addLayer(vacuum_layer)
    return multi_layer


def run_simulation(det, ff, trafo=None):
    """Run simulation and plot results

    :param det: Detector limits
    :param ff: Form factor
    :param trafo: Optional rotation
    """
    zero = ba.cvector_t(0, 0, 0)
    volume = abs(ff.evaluate_for_q(zero))
    print("Volume: %g" % volume)
    sample = get_sample(ff, trafo)
    beam = ba.Beam(1, 1*angstrom, ba.Direction(1e-8*deg, 0))
    simulation = ba.GISASSimulation(beam, sample, det)
    simulation.runSimulation()
    result = simulation.result()
    #import ba_plot
    #ba_plot.plot_simulation_result(result)
    data = result.array()
    nor = data[det.axis(1).size() - det.axis(1).findClosestIndex(0) - 1,
               det.axis(0).findClosestIndex(0)]
    data /= nor
    return data + 1e-80  # for log scale



#!/usr/bin/env python

"""

"""

import glob, re, sys
from io import open

if len(sys.argv)<3:
    print( """
Usage: count-line-length <max_mum_chars_per_line> <files>

Returns 1 if any given file
    - exceeds the given number of characters per line,
    - or contains a tab.
Returns 0 otherwise.
""" )
    sys.exit(-1)

fworst = ""
lworst = 0
totlin = 0
totexc = 0
tottab = 0
nftot = 0
nfexc = 0
nftab = 0

limit = int(sys.argv[1])
flist = sys.argv[2:]
for fn in flist:
    # read in
    fd = open( fn, 'r', encoding="utf-8" )
    txt = fd.read().split( '\n' )
    fd.close

    n = len(txt)
    maxlen = 0
    maxwhere = -1
    nexclin = 0
    ntablin = 0
    for i in range(n):
        lin = txt[i]
        if re.match(r'\s*\/\/', lin):
            continue # ignore comment lines
        if '\t' in lin:
            ntablin += 1
            lin = lin.replace( '\t', '    ' )
        l = len(lin)
        if l>maxlen:
            maxlen = l
            maxwhere = i
        if l>limit:
            nexclin += 1

    if nexclin>0:
        print( "%s: %i/%i lines too long, line %i longest with %i chars" %
               (fn, nexclin, n, maxwhere+1, maxlen) )
        totexc += nexclin
        nfexc += 1
    if ntablin>0:
        print( "%s: %i/%i lines contain a TAB" % (fn, ntablin, n) )
        tottab += ntablin
        nftab += 1
    totlin += n
    nftot += 1
    if maxlen>lworst:
        lworst = maxlen
        fworst = fn

exitval = 0
if nfexc>0:
    print( "total: %i/%i lines too long, in %i/%i files, longest in %s with %i chars" %
           (totexc, totlin, nfexc, nftot, fworst, lworst) )
    exitval = 1
if nftab>0:
    print( "total: %i/%i lines contain a TAB, in %i/%i files" %
           (tottab, totlin, nftab, nftot) )
    exitval = 1

sys.exit(exitval)



#!/usr/bin/env python3
"""
Plot form factor.
"""
import bornagain as ba
from bornagain import nm, deg
import bornplot as bp

det = ba.SphericalDetector(200, 5*deg, 2.5*deg, 2.5*deg)
n = 1
results = []
for i in range(n):
    ff = ba.FormFactorFullSphere(3.9*nm)
    data = bp.run_simulation(det, ff)
    results.append(bp.Result(i, data))

bp.make_plot(results, det, "ff_FullSphere")



"""
Collection of utils for testing
"""

import gzip, numpy, sys, os
import bornagain as ba
from bornagain import deg, angstrom

REFERENCE_DIR = "@TEST_REFERENCE_DIR_PY_CORE@"


def get_difference(data, reference):
    """
    calculate numeric difference between result and reference data
    """
    reference = reference.reshape(data.shape)
    data -= reference
    diff = 0.0
    epsilon = sys.float_info.epsilon
    for x, y in numpy.ndindex(data.shape):
        v1 = data[x][y]
        v2 = reference[x][y]
        if v1 <= epsilon and v2 <= epsilon:
            diff += 0.0
        elif v2 <= epsilon:
            diff += abs(v1/epsilon)
        else:
            diff += abs(v1/v2)
    diff = diff/data.size
    if numpy.isnan(diff):
        raise ba.Exception("get_difference", "isnan")
    return diff


def get_reference_data(filename):
    """
    read and return reference data from file
    """
    return ba.IntensityDataIOFactory.readIntensityData(
        os.path.join(REFERENCE_DIR, filename))


def get_simulation_MiniGISAS(sample):
    detector = ba.SphericalDetector(25, -2*deg, 2*deg, 25, 0, 2*deg)
    beam = ba.Beam(1., 1*angstrom, ba.Direction(0.2*deg, 0))
    return ba.GISASSimulation(beam, sample, detector)



#!/usr/bin/env python3
"""
Plot form factors.
"""
import bornagain as ba
from bornagain import nm, deg
import bornplot as bp
import math

det = ba.SphericalDetector(200, 10*deg, 0, 0)
n = 3
results = []
edge = 3.2

title = 'face normal'
trafo = ba.RotationY(26.5651*deg)
ff = ba.FormFactorDodecahedron(edge*nm)
data = bp.run_simulation(det, ff, trafo)
results.append(bp.Result(0, data, title))

title = 'vertex normal'
trafo = ba.RotationY(-52.6226*deg)
ff = ba.FormFactorDodecahedron(edge*nm)
data = bp.run_simulation(det, ff, trafo)
results.append(bp.Result(1, data, title))

title = 'edge normal'
trafo = ba.RotationY(58.2825*deg)
ff = ba.FormFactorDodecahedron(edge*nm)
data = bp.run_simulation(det, ff, trafo)
results.append(bp.Result(2, data, title))

bp.make_plot(results, det, "ff_Dodecahedron_sym")



#!/usr/bin/env python3
"""
MultiLayer with correlated roughness
"""
import bornagain as ba
from bornagain import deg, nm


def get_sample():
    """
    Returns a sample with two layers on a substrate, with correlated roughnesses.
    """
    # defining materials
    m_vacuum = ba.HomogeneousMaterial("ambience", 0, 0)
    m_part_a = ba.HomogeneousMaterial("PartA", 5e-6, 0)
    m_part_b = ba.HomogeneousMaterial("PartB", 10e-6, 0)
    m_substrate = ba.HomogeneousMaterial("substrate", 15e-6, 0)

    # defining layers
    l_ambience = ba.Layer(m_vacuum)
    l_part_a = ba.Layer(m_part_a, 2.5*nm)
    l_part_b = ba.Layer(m_part_b, 5*nm)
    l_substrate = ba.Layer(m_substrate)

    roughness = ba.LayerRoughness()
    roughness.setSigma(1*nm)
    roughness.setHurstParameter(0.3)
    roughness.setLatteralCorrLength(5*nm)

    my_sample = ba.MultiLayer()

    # adding layers
    my_sample.addLayer(l_ambience)

    n_repetitions = 5
    for i in range(n_repetitions):
        my_sample.addLayerWithTopRoughness(l_part_a, roughness)
        my_sample.addLayerWithTopRoughness(l_part_b, roughness)

    my_sample.addLayerWithTopRoughness(l_substrate, roughness)
    my_sample.setCrossCorrLength(10*nm)

    return my_sample


def get_simulation(sample):
    beam = ba.Beam(5e11, 0.1*nm, ba.Direction(0.2*deg, 0))
    detector = ba.SphericalDetector(200, 1*deg, 0, 0.5*deg)
    simulation = ba.GISASSimulation(beam, sample, detector)
    return simulation


if __name__ == '__main__':
    import ba_plot
    sample = get_sample()
    simulation = get_simulation(sample)
    ba_plot.run_and_plot(simulation)



import numpy, os, sys, unittest
import bornagain as ba
from bornagain import nm


class ParameterPoolTest(unittest.TestCase):
    def test_parameterPoolAccess(self):
        """
        Checks values in particle's parameter pool
        """
        ff = ba.FormFactorCylinder(5*nm, 6*nm)
        particle = ba.Particle(ba.HomogeneousMaterial("Vacuum", 0, 0), ff)
        particle.setAbundance(1)
        particle.setPosition(2, 3, 4)

        # print(particle.parametersToString())
        # print(particle.treeToString())

        pool = particle.parameterPool()
        self.assertEqual(pool.size(), 4)
        self.assertEqual(
            pool.parameterNames(),
            ('Abundance', 'PositionX', 'PositionY', 'PositionZ'))

        expected = {
            'Abundance': 1,
            'PositionX': 2,
            'PositionY': 3,
            'PositionZ': 4.0
        }
        for par in pool:
            print(par.value(), par.getName(), par.limits().toString())
            self.assertEqual(par.value(), expected[par.getName()])

    def test_parameterPoolModify(self):
        """
        Modification of particle parameters via parameter pool
        """
        ff = ba.FormFactorCylinder(5*nm, 6*nm)
        particle = ba.Particle(ba.HomogeneousMaterial("Vacuum", 0, 0), ff)
        particle.setAbundance(1)
        particle.setPosition(2, 3, 4)

        pool = particle.parameterPool()
        pool.setParameterValue('Abundance', 10)
        pool[1].setValue(20)  # PositionX
        pool.parameter('PositionY').setValue(30)

        expected = {
            'Abundance': 10,
            'PositionX': 20,
            'PositionY': 30,
            'PositionZ': 4.0
        }

        for par in pool:
            self.assertEqual(par.value(), expected[par.getName()])

    def test_parameterTreeAccess(self):
        """
        Checks values in particle's parameter tree. Parameter tree is a pool with parameters of
        particle and its children (in given case, form factor of cylinder)
        """
        ff = ba.FormFactorCylinder(5*nm, 6*nm)
        particle = ba.Particle(ba.HomogeneousMaterial("Vacuum", 0, 0), ff)
        particle.setAbundance(1)
        particle.setPosition(2, 3, 4)

        pool = particle.createParameterTree()

        expected = {
            '/Particle/Abundance': 1,
            '/Particle/PositionX': 2,
            '/Particle/PositionY': 3,
            '/Particle/PositionZ': 4,
            '/Particle/Cylinder/Radius': 5,
            '/Particle/Cylinder/Height': 6.0
        }

        for par in pool:
            self.assertEqual(par.value(), expected[par.getName()])

    def test_parameterTreeModify(self):
        """
        Modifies values of particle's parameter tree.
        """
        ff = ba.FormFactorCylinder(5*nm, 6*nm)
        particle = ba.Particle(ba.HomogeneousMaterial("Vacuum", 0, 0), ff)
        particle.setAbundance(1)
        particle.setPosition(2, 3, 4)

        # print(particle.parametersToString())
        # print(particle.treeToString())

        pool = particle.createParameterTree()
        pool.setParameterValue('/Particle/Abundance', 10)
        pool[1].setValue(20)  # PositionX
        pool.parameter('/Particle/PositionY').setValue(30)
        pool.setMatchedParametersValue('*Cylinder*', 50)

        expected = {
            '/Particle/Abundance': 10,
            '/Particle/PositionX': 20,
            '/Particle/PositionY': 30,
            '/Particle/PositionZ': 4,
            '/Particle/Cylinder/Radius': 50,
            '/Particle/Cylinder/Height': 50.0
        }

        for par in pool:
            self.assertEqual(par.value(), expected[par.getName()])

    def test_parameterModify(self):
        """
        Modification of particle's parameters without intermediate access to parameter pool
        """
        ff = ba.FormFactorCylinder(5*nm, 6*nm)
        particle = ba.Particle(ba.HomogeneousMaterial("Vacuum", 0, 0), ff)
        particle.setAbundance(1)
        particle.setPosition(2, 3, 4)

        particle.setParameterValue('/Particle/Abundance', 10)
        particle.setParameterValue('PositionZ', 40)
        particle.setParameterValue('*Cylinder*', 50)

        self.assertEqual(particle.abundance(), 10)
        self.assertEqual(particle.position().z(), 40)


if __name__ == '__main__':
    unittest.main()



"""
The script compares BornAgain/ThirdParty/RootMinimizers with original ROOT installation to propagate code modifications
from ROOT to our source tree.
"""
import os
import filecmp
import shutil

ROOT_SOURCE = "/home/pospelov/software/source/root"
BORNAGAIN_SOURCE = "/home/pospelov/development/BornAgain/BornAgain/ThirdParty/Fit/RootMinimizers"


def copy_file_to_file(source, destination):
    """
    Copies file to another.
    """
    shutil.copyfile(source, destination)


def find_files(rootdir):
    """
    Yield recursive list of files in given rootdir
    """
    for subdir, dirs, files in os.walk(rootdir):
        for filename in files:
            #yield os.sep.join([subdir, filename])
            yield subdir, filename


def get_source_files(rootdir):
    """
    Return list of *.cxx *.cpp *.h files in given rootdir
    """
    extensions = [".h", ".cpp", ".cxx"]
    result = []
    for subdir, filename in find_files(rootdir):
            name, ext = os.path.splitext(filename)
            if ext in extensions:
                result.append(os.sep.join([subdir, filename]))
    return result


#def find_files_with_same_name(rootdir, name_to_find):
#    """
#    Returns recursive list of files in rootdir with given name
#    """
#    result = []
#    for file in find_files(rootdir):
#        if os.path.basename(file) == name_to_find:
#            result.append(file)
#    return result


def find_files_with_same_name(filename_list, name_to_find):
    same_names = []
    for filename in filename_list:
        if os.path.basename(filename) == name_to_find:
            same_names.append(filename)

    result = []
    if len(same_names) > 1:
        for name in same_names:
            if "Math" in name or "TMVA" in name:
                result.append(name)
    else:
        result = same_names

    return result


def build_difference_map():
    ba_files = get_source_files(BORNAGAIN_SOURCE)
    root_files = get_source_files(ROOT_SOURCE)

    for ba_file in ba_files:
        found_files = find_files_with_same_name(root_files, os.path.basename(ba_file))
        if len(found_files) == 1:
            if filecmp.cmp(ba_file, found_files[0]) == False:
                cmd = "cp " + found_files[0] + " " + os.path.dirname(ba_file)
                print(len(found_files), found_files[0], ba_file)
                copy_file_to_file(found_files[0], ba_file)
        else:
            print(len(found_files), found_files, ba_file)


if __name__ == '__main__':
    build_difference_map()



"""
Test of rotation/position of box in the middle layer of 3 layers system.
FormFactorBox(50, 20, 10), is rotated around Z, then around Y, then shifted up and then
compared with FormFactorBox(10, 50, 20)
"""

from __future__ import print_function
import os, sys, unittest

import utils
import bornagain as ba
from bornagain import deg, kvector_t

layer_thickness = 100


class BoxTransformationsTest(unittest.TestCase):
    def get_sample(self, particle):
        mAmbience = ba.HomogeneousMaterial("Vacuum", 0, 0)
        mMiddle = ba.HomogeneousMaterial("Teflon", 2.900e-6, 6.019e-9)
        mSubstrate = ba.HomogeneousMaterial("Substrate", 3.212e-6,
                                            3.244e-8)

        layout = ba.ParticleLayout()
        layout.addParticle(particle)

        vacuum_layer = ba.Layer(mAmbience)
        middle_layer = ba.Layer(mMiddle, layer_thickness)
        middle_layer.addLayout(layout)
        substrate = ba.Layer(mSubstrate)

        multi_layer = ba.MultiLayer()
        multi_layer.addLayer(vacuum_layer)
        multi_layer.addLayer(middle_layer)
        multi_layer.addLayer(substrate)
        return multi_layer

    def get_result(self, particle):
        sample = self.get_sample(particle)
        simulation = utils.get_simulation_MiniGISAS(sample)
        simulation.runSimulation()
        return simulation.result()

    def testBoxTransform(self):
        """
        Reference box of (10,50,20) size is compared against the box (50,20,10) with two rotations applied to get
        reference one
        """
        mParticle = ba.HomogeneousMaterial("Ag", 1.245e-5, 5.419e-7)

        # reference box
        length = 10
        width = 50
        height = 20

        box = ba.Particle(mParticle,
                          ba.FormFactorBox(length, width, height))
        box.setPosition(kvector_t(0, 0, -layer_thickness/2 - height/2))
        reference_data = self.get_result(box)
        #IntensityDataIOFactory.writeIntensityData(reference_data, "ref_TransformBox.int")

        # second box
        length = 50
        width = 20
        height = 10
        box = ba.Particle(mParticle,
                          ba.FormFactorBox(length, width, height))
        box.setRotation(ba.RotationZ(90*deg))
        box.rotate(ba.RotationY(90*deg))
        box.setPosition(kvector_t(0, 0, -layer_thickness/2))

        data = self.get_result(box)

        diff = ba.RelativeDifference(data, reference_data)
        print(diff)
        self.assertLess(diff, 1e-10)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
An example of computing reflectivity on a
magnetized sample.
"""

import bornagain as ba
from bornagain import angstrom, deg, nm, nm2, kvector_t

import matplotlib.pyplot as plt


def get_sample():
    """
    Defines sample and returns it
    """

    # Define materials
    material_Ambient = ba.MaterialBySLD("Ambient", 0, 0)
    magnetic_field = kvector_t(0, 1e8, 0)
    material_Layer = ba.MaterialBySLD("Layer", 0.0001, 1e-08,
                                      magnetic_field)
    material_Substrate = ba.MaterialBySLD("Substrate", 7e-05, 2e-06)

    # Define layers
    layer_1 = ba.Layer(material_Ambient)
    layer_2 = ba.Layer(material_Layer, 10*nm)
    layer_3 = ba.Layer(material_Substrate)

    # Define sample
    sample = ba.MultiLayer()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)
    sample.addLayer(layer_3)

    return sample


def get_simulation(sample, scan_size=500):
    """
    Defines and returns a specular simulation.
    """
    simulation = ba.SpecularSimulation()
    scan = ba.AngularSpecScan(1.54*angstrom, scan_size, 0, 5*deg)
    simulation.setScan(scan)
    simulation.setSample(sample)
    return simulation


def run_simulation(polarization=ba.kvector_t(0, 1, 0),
                   analyzer=ba.kvector_t(0, 1, 0)):
    """
    Runs simulation and returns its result.
    """
    sample = get_sample()
    simulation = get_simulation(sample)

    # adding polarization and analyzer operator
    simulation.beam().setPolarization(polarization)
    simulation.detector().setAnalyzerProperties(analyzer, 1, 0.5)

    simulation.runSimulation()
    return simulation.result()


def plot(data, labels):

    plt.figure()
    for d, l in zip(data, labels):
        plt.semilogy(d.axis(), d.array(), label=l, linewidth=1)

    plt.legend(loc='upper right')
    plt.gca().yaxis.set_ticks_position('both')
    plt.gca().xaxis.set_ticks_position('both')

    plt.xlabel(r"$\alpha_i$ [deg]")
    plt.ylabel("Reflectivity")

    plt.tight_layout()
    plt.show()


if __name__ == '__main__':
    results_pp = run_simulation(ba.kvector_t(0, 1, 0),
                                ba.kvector_t(0, 1, 0))
    results_mm = run_simulation(ba.kvector_t(0, -1, 0),
                                ba.kvector_t(0, -1, 0))
    plot([results_pp, results_mm], ["$++$", "$--$"])



"""
Functional test to check slicing mechanism for particle compositions
when they are crossing an interface.
"""

from __future__ import print_function
import os, sys, unittest

import utils
import bornagain as ba
from bornagain import deg, kvector_t, nm

mSubstrate = ba.HomogeneousMaterial("Substrate", 3.212e-6, 3.244e-8)
mAmbience = ba.HomogeneousMaterial("Vacuum", 0, 0)
mParticle = ba.HomogeneousMaterial("Ag", 1.245e-5, 5.419e-7)
sphere_radius = 10.0
bottom_cup_height = 4.0


class SlicedSpheresTest(unittest.TestCase):
    def get_sample(self, particle_to_air=None, particle_to_substrate=None):
        """
        Helper function returning a multilayer (air, substrate) using particles provided
        by the user.
        """

        vacuum_layer = ba.Layer(mAmbience)
        if particle_to_air:
            layout = ba.ParticleLayout()
            layout.addParticle(particle_to_air)
            vacuum_layer.addLayout(layout)

        substrate = ba.Layer(mSubstrate)
        if particle_to_substrate:
            layout = ba.ParticleLayout()
            layout.addParticle(particle_to_substrate)
            substrate.addLayout(layout)

        multi_layer = ba.MultiLayer()
        multi_layer.addLayer(vacuum_layer)
        multi_layer.addLayer(substrate)
        return multi_layer

    def get_result(self, particle_to_air=None, particle_to_substrate=None):
        sample = self.get_sample(particle_to_air, particle_to_substrate)
        simulation = utils.get_simulation_MiniGISAS(sample)
        simulation.runSimulation()
        return simulation.result()

    def get_composition(self, top_material, bottom_material):
        """
        Returns particle composition representing sphere made of two different materials.
        Origin of new object is at the bottom of resulting sphere.
        """

        topCup = ba.Particle(
            top_material,
            ba.FormFactorTruncatedSphere(
                sphere_radius, sphere_radius*2 - bottom_cup_height, 0))
        bottomCup = ba.Particle(
            bottom_material,
            ba.FormFactorTruncatedSphere(
                sphere_radius, sphere_radius*2,
                sphere_radius*2 - bottom_cup_height))

        #  origin of resulting sphere will be at the bottom
        result = ba.ParticleComposition()
        result.addParticle(topCup, kvector_t(0, 0, bottom_cup_height))
        result.addParticle(bottomCup, kvector_t(0, 0, 0))

        return result

    def get_composition_v2(self, top_material, bottom_material):
        """
        Returns particle composition representing sphere made of two different materials.
        Alternative to previous method.
        Rotation is used to get bottom part
        """

        topCup = ba.Particle(
            top_material,
            ba.FormFactorTruncatedSphere(
                sphere_radius, sphere_radius*2 - bottom_cup_height, 0))
        bottomCup = ba.Particle(
            bottom_material,
            ba.FormFactorTruncatedSphere(sphere_radius, bottom_cup_height,
                                         0))
        bottomCup.setRotation(ba.RotationX(180*deg))

        #  origin of resulting sphere will be at the bottom
        result = ba.ParticleComposition()
        result.addParticle(topCup, kvector_t(0, 0, bottom_cup_height))
        result.addParticle(bottomCup, kvector_t(0, 0, bottom_cup_height))

        return result

    def testComposition(self):
        """
        Compares two simulation intended to  provide identical results.
        Simulation #1: spherical particle on top of substrate
        Simulation #2: spherical composition on top of substrate, where top and bottom are made of same material
        """

        # spherical particle
        sphere = ba.Particle(mParticle,
                             ba.FormFactorFullSphere(sphere_radius))
        reference = self.get_result(sphere)

        # spherical composition
        composition = self.get_composition(mParticle, mParticle)
        data = self.get_result(composition)

        diff = ba.RelativeDifference(data, reference)
        print(diff)
        self.assertLess(diff, 1e-10)

    def testCompositionBuilds(self):
        """
        Compares two simulation intended to  provide identical results.
        Simulation #1: spherical particle composition on top of substrate
        Simulation #2: same, but made using rotation
        """

        composition = self.get_composition(mParticle, mSubstrate)
        reference = self.get_result(composition)

        # spherical composition
        composition2 = self.get_composition_v2(mParticle, mSubstrate)
        data = self.get_result(composition2)

        diff = ba.RelativeDifference(data, reference)
        print(diff)
        self.assertLess(diff, 1e-10)

    def testInvisibleComposition(self):
        """
        Compares two simulation intended to  provide identical results.
        Simulation #1: no particles at all
        Simulation #2: spherical composition on top of substrate crossing the interface
        Bottom part of composition is made of substrate material, top part from vacuum layer material
        """

        shift = bottom_cup_height

        # Scattering from empty multilayer
        reference = self.get_result()

        # spherical composition
        composition = self.get_composition(mAmbience, mSubstrate)
        composition.setPosition(0, 0, -shift)
        data = self.get_result(composition)

        diff = ba.RelativeDifference(data, reference)
        print(diff)
        self.assertLess(diff, 1e-10)

    def testSlicedComposition(self):
        """
        Compares two simulation intended to  provide identical results.
        Simulation #1: spherical particle on top of substrate
        Simulation #2: spherical composition on top of substrate, where top and bottom are made of same material
        Both particles are inserted in vacuum layer with shift to go below interface
        """

        shift = 3*nm

        # spherical particle
        sphere = ba.Particle(mParticle,
                             ba.FormFactorFullSphere(sphere_radius))
        sphere.setPosition(0, 0, -shift)
        reference = self.get_result(sphere)

        # spherical composition
        composition = self.get_composition(mParticle, mParticle)
        composition.setPosition(0, 0, -shift)
        data = self.get_result(composition)

        diff = ba.RelativeDifference(data, reference)
        print(diff)
        self.assertLess(diff, 1e-10)

    def testSphericalCupOnTopOfSubstrate(self):
        """
        Compares two simulation intended to  provide identical results.
        Simulation #1: truncated sphere on top of substrate.
        Simulation #2: spherical particle composition crossing the interface.
        Bottom part of composition is made from substrate material.
        both cases should look like a truncated sphere on top of substrate.
        """

        # truncated sphere on top of substrate with height 16nm
        truncatedSphere = ba.Particle(
            mParticle,
            ba.FormFactorTruncatedSphere(
                sphere_radius, sphere_radius*2 - bottom_cup_height, 0))
        reference = self.get_result(truncatedSphere)

        # Particle composition, top part made of same material, as particle. Bottom part made of same material as substrate.
        composition = self.get_composition(mParticle, mSubstrate)

        composition_shift = bottom_cup_height
        composition.setPosition(0, 0, -composition_shift)
        data = self.get_result(composition)

        diff = ba.RelativeDifference(data, reference)
        print(diff)
        self.assertLess(diff, 1e-10)


if __name__ == '__main__':
    unittest.main()



import matplotlib.pyplot as plt
import seaborn as sns
from datetime import datetime
from .file_types import FileTypes, filetype


selected_file_type = [FileTypes.CORE, FileTypes.FTEST, FileTypes.UTEST, FileTypes.GUI, FileTypes.PYAPI]


def read_history(filename):
    xvals = []  # Time
    ydata = {}  # id of file type .vs. LOC .vs. time
    for x in selected_file_type:
        ydata[x] = []

    print("Reading file {0}".format(filename))
    with open(filename, 'r') as the_file:
        for line in the_file:
            parts = line.strip().split()
            date = datetime.strptime(parts[0] + " " + parts[1], '%Y-%m-%d %H:%M:%S')
            xvals.append(date)
            for x in selected_file_type:
                ydata[x].append(int(parts[2+x]))

    yvals = []
    descr = []
    for key in ydata:
        descr.append(FileTypes.descr[key])
        yvals.append(ydata[key])

    # printing summary of LOC
    for x in range(0, len(yvals)):
        print("{:18} : {:10}".format(descr[x], yvals[x][-1]))

    return xvals, yvals, descr


def history_plot(filename):
    xvals, yvals, descr = read_history(filename)

    # figure size
    my_dpi = 96
    plt.figure(figsize=(1600*1.2 / my_dpi, 900*1.2 / my_dpi), dpi=my_dpi)
    plt.style.use('seaborn-bright')

    # making stackplot
    plt.stackplot(xvals, yvals)
    pal = ["#3399ff", "#ffcc00", "#ff0000", "#0033ff", "#999999"]
    plt.stackplot(xvals, yvals, labels=descr, colors=pal)

    # styling axes and grid
    plt.grid(color='gray', linestyle='dashed')
    ax = plt.gca()
    ax.set_axisbelow(True)
    plt.ylim(0.0, 220e+03)
    plt.tick_params(axis='both', which='major', labelsize=14)

    # making inverse legend
    handles, labels = ax.get_legend_handles_labels()
    ax.legend(handles[::-1], labels[::-1], loc='upper left', prop={'size': 18})

    # saving plot
    plt.savefig('lines_of_code.png', dpi=my_dpi, bbox_inches='tight')



"""
Test of rotation/positioning of simple cubic particle. Original particle is compared with the one obtained
"""

from __future__ import print_function
import os, sys, unittest

import utils
from bornagain import *


class RotationsCubeTest(unittest.TestCase):
    """
    Test of rotations and translations of simple cube in three layers system
    """
    def get_sample(self,
                   formfactor,
                   rot=None,
                   pos=None,
                   layout_rot=None,
                   layout_pos=None,
                   add_to="Vacuum"):
        mAmbience = HomogeneousMaterial("Vacuum", 0, 0)
        mParticle = HomogeneousMaterial("Particle", 6e-4, 2e-8)
        mMiddle = HomogeneousMaterial("MidleLayer", 5e-5, 2e-8)
        mSubstrate = HomogeneousMaterial("Substrate", 6e-6, 2e-8)

        particle = Particle(mParticle, formfactor)
        if pos:
            particle.setPosition(pos)
        if rot:
            particle.setRotation(rot)

        layout = ParticleLayout()
        if layout_rot and layout_pos:
            layout.addParticle(particle, 1, layout_pos, layout_rot)
        elif layout_rot and not layout_pos:
            layout.addParticle(particle, 1, kvector_t(0, 0, 0), layout_rot)
        elif not layout_rot and layout_pos:
            layout.addParticle(particle, 1, layout_pos)
        else:
            layout.addParticle(particle)

        vacuum_layer = Layer(mAmbience)
        middle_layer = Layer(mSubstrate, 50)
        substrate = Layer(mSubstrate)

        if add_to == "Vacuum":
            vacuum_layer.addLayout(layout)
        else:
            middle_layer.addLayout(layout)

        multi_layer = MultiLayer()
        multi_layer.addLayer(vacuum_layer)
        multi_layer.addLayer(middle_layer)
        multi_layer.addLayer(substrate)
        return multi_layer

    def get_result(self, data, add_to="Vacuum"):
        ff = data[0]
        rot = data[1]
        pos = data[2]
        layout_rot = data[3]
        layout_pos = data[4]
        sample = self.get_sample(ff, rot, pos, layout_rot, layout_pos,
                                 add_to)
        # simulation = self.get_simulation(sample)
        simulation = utils.get_simulation_MiniGISAS(sample)
        simulation.runSimulation()
        return simulation.result()

    def get_difference(self, reference_data, test_data, add_to="Vacuum"):
        intensity = self.get_result(test_data, add_to)
        return RelativeDifference(reference_data, intensity)

    def testRotationZ(self):
        """
        Cube is Z-rotated either through setRotation method or through particle layout. The result is compared
        with unrotated cube.
        """
        box = FormFactorBox(10, 10, 10)

        data_to_test = [
            # ff  rot                     pos                    layout_rot              layout_pos
            (box, None, None, None, None),  # reference
            (box, RotationZ(90.*deg), None, None,
             None),  # rotating particle
            (box, RotationZ(-90.*deg), None, None, None),
            (box, RotationZ(180.*deg), None, None, None),
            (box, None, None, RotationZ(90.*deg),
             None),  # rotating through layout
            (box, RotationZ(45.*deg), None, RotationZ(45.*deg),
             None),  # cumulative rotation
        ]

        reference_data = self.get_result(data_to_test[0])

        isSuccess = True
        for i in range(1, len(data_to_test)):
            diff = self.get_difference(reference_data, data_to_test[i])
            print("{0}    #{1}   diff {2:.2e}".format(
                self.testRotationZ.__name__, i, diff))
            if (diff > 1e-10): isSuccess = False

        self.assertTrue(isSuccess)

    def testRotationY(self):
        """
        Cube is Y-rotated either through setRotation method or through particle layout. Additional translation is
        applied if necessary. The result is compared with unrotated cube.
        """
        box = FormFactorBox(10, 10, 10)

        data_to_test = [
            # ff  rot                     pos                    layout_rot              layout_pos
            (box, None, None, None, None),  # reference
            (box, RotationY(90.*deg), kvector_t(0, 0, 5), None,
             None),  # rotating and translating
            (box, None, None, RotationY(90.*deg),
             kvector_t(0, 0, 5)),  # rotating and translating
            (box, RotationY(90.*deg), None, None,
             kvector_t(0, 0, 5)),  # rotating and translating
            (box, RotationY(45.*deg), kvector_t(0, 0,
                                                0), RotationY(45.*deg),
             kvector_t(0, 0, 5)),  # rotating and translating
        ]

        reference_data = self.get_result(data_to_test[0])

        isSuccess = True
        for i in range(1, len(data_to_test)):
            diff = self.get_difference(reference_data, data_to_test[i])
            print("{0}    #{1}   diff {2:.2e}".format(
                self.testRotationY.__name__, i, diff))
            if (diff > 1e-10): isSuccess = False

        self.assertTrue(isSuccess)

    def testRotationX(self):
        """
        Cube is Z-rotated either through setRotation method or through particle layout. Additional translation is
        applied if necessary. The result is compared with unrotated cube.
        """
        box = FormFactorBox(10, 10, 10)

        data_to_test = [
            # ff  rot                     pos                    layout_rot              layout_pos
            (box, None, None, None, None),  # reference
            (box, RotationX(90.*deg), kvector_t(0, 0, 5), None,
             None),  # rotating and translating
            (box, None, None, RotationX(90.*deg),
             kvector_t(0, 0, 5)),  # rotating and translating
            (box, RotationX(90.*deg), None, None,
             kvector_t(0, 0, 5)),  # rotating and translating
            (box, RotationX(45.*deg), kvector_t(0, 0,
                                                0), RotationX(45.*deg),
             kvector_t(0, 0, 5)),  # rotating and translating
        ]

        reference_data = self.get_result(data_to_test[0])

        isSuccess = True
        for i in range(1, len(data_to_test)):
            diff = self.get_difference(reference_data, data_to_test[i])
            print("{0}    #{1}   diff {2:.2e}".format(
                self.testRotationX.__name__, i, diff))
            if (diff > 1e-10): isSuccess = False

        self.assertTrue(isSuccess)

    def testRotationsInMiddleLayer(self):
        """
        """
        box = FormFactorBox(10, 10, 10)

        data_to_test = [
            # ff  rot                     pos                    layout_rot              layout_pos
            (box, None, kvector_t(0, 0, -25), None, None),  # reference
            (box, RotationX(90.*deg), kvector_t(0, 0, -20), None,
             None),  # rotating and translating
        ]

        reference_data = self.get_result(data_to_test[0], "add_to_middle")

        isSuccess = True
        for i in range(1, len(data_to_test)):
            diff = self.get_difference(reference_data, data_to_test[i],
                                       "add_to_middle")
            print("{0}    #{1}   diff {2:.2e}".format(
                self.testRotationX.__name__, i, diff))
            if (diff > 1e-10): isSuccess = False

        self.assertTrue(isSuccess)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
Plots data stored in BornAgain's .int or .int.gz format.
Can handle both 1D and 2D arrays.
"""

import argparse
import sys
import numpy as np
import bornagain as ba
from matplotlib import pyplot as plt
from matplotlib import rc, colors
rc('font', **{'family': 'sans-serif', 'sans-serif': ['Helvetica']})
rc('text', usetex=True)


def plot_int(args):
    data = ba.IntensityDataIOFactory.readIntensityData(args.file)
    if args.max:
        intensity_max = args.max
    else:
        intensity_max = data.getMaximum()
    if args.min:
        intensity_min = args.min
    else:
        intensity_min = data.getMinimum()
    ylog = not args.ylin
    if ylog and intensity_min <= 0:
        intensity_min = intensity_max/1e6

    if args.verbose:
        print(f'Data extend from {data.getMinimum()} to {data.getMaximum()}')
    if data.rank() == 1:
        # some white space above and below automatically determined y range:
        if not args.max:
            if args.ylin:
                intensity_max = intensity_max + (intensity_max -
                                                 intensity_min)*0.05
            else:
                intensity_max = intensity_max*(intensity_max/
                                               intensity_min)**0.05
        if not args.min:
            if args.ylin:
                intensity_min = intensity_min - (intensity_max -
                                                 intensity_min)*0.05
            else:
                intensity_min = intensity_min/(intensity_max/
                                               intensity_min)**0.05
        plot_int_1d(data, ylog, intensity_min, intensity_max)
    elif data.rank() == 2:
        plot_int_2d(data, ylog, intensity_min, intensity_max)
    else:
        exit("Error in plot_int: wrong data rank")


def plot_int_2d(histogram, ylog, intensity_min, intensity_max):
    plot_raw_data_2d(histogram.array(), [
        histogram.getXmin()/ba.deg,
        histogram.getXmax()/ba.deg,
        histogram.getYmin()/ba.deg,
        histogram.getYmax()/ba.deg
    ], ylog, intensity_min, intensity_max)


def plot_raw_data_2d(values, extent_array, ylog, intensity_min, intensity_max):
    if ylog:
        norm = colors.LogNorm(intensity_min, intensity_max)
    else:
        norm = colors.Normalize(intensity_min, intensity_max)
    im = plt.imshow(values, norm=norm, extent=extent_array, aspect='auto')
    cb = plt.colorbar(im)
    cb.set_label(r'Intensity (arb. u.)', size=16)
    plt.xlabel(r'$\phi_f (^{\circ})$', fontsize=16)
    plt.ylabel(r'$\alpha_f (^{\circ})$', fontsize=16)
    plt.show()


def plot_int_1d(histogram, ylog, intensity_min, intensity_max):
    axis_values = np.asarray(histogram.xAxis().binCenters())/ba.deg
    array_values = histogram.array()
    plt.ylim(intensity_min, intensity_max)
    plot_raw_data_1d(axis_values, array_values, ylog)


def plot_raw_data_1d(axis, values, ylog):
    if ylog:
        plt.semilogy(axis, values)
    else:
        plt.plot(axis, values)

    plt.xlabel(r'$\alpha_i (^{\circ})$', fontsize=16)
    plt.ylabel(r'Value (a.u.)', fontsize=16)
    plt.show()


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-v",
                        "--verbose",
                        action="store_true",
                        help="print some output to the terminal")
    parser.add_argument("file",
                        type=str,
                        help="input data file (.int or .int.gz)")
    parser.add_argument("-l", "--min", type=float, help="upper plot limit")
    parser.add_argument("-u", "--max", type=float, help="lower plot limit")
    parser.add_argument(
        "-y",
        "--ylin",
        action="store_true",
        help="linear y scale",
    )
    args = parser.parse_args()

    plot_int(args)



#!/usr/bin/env python3
"""
Fitting example: simultaneous fit of two datasets
"""

import numpy as np
import matplotlib
from matplotlib import pyplot as plt
import bornagain as ba
from bornagain import deg, nm
import ba_plot as bp

def get_sample(params):
    """
    Returns a sample with uncorrelated cylinders and pyramids.
    """
    radius_a = params["radius_a"]
    radius_b = params["radius_b"]
    height = params["height"]

    m_vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)
    m_substrate = ba.HomogeneousMaterial("Substrate", 6e-6, 2e-8)
    m_particle = ba.HomogeneousMaterial("Particle", 6e-4, 2e-8)

    formfactor = ba.FormFactorHemiEllipsoid(radius_a, radius_b, height)
    particle = ba.Particle(m_particle, formfactor)

    layout = ba.ParticleLayout()
    layout.addParticle(particle)

    vacuum_layer = ba.Layer(m_vacuum)
    vacuum_layer.addLayout(layout)

    substrate_layer = ba.Layer(m_substrate)
    multi_layer = ba.MultiLayer()
    multi_layer.addLayer(vacuum_layer)
    multi_layer.addLayer(substrate_layer)
    return multi_layer


def get_simulation(params):
    """
    Returns a GISAXS simulation with beam and detector defined.
    """
    incident_angle = params["incident_angle"]

    simulation = ba.GISASSimulation()
    simulation.setDetectorParameters(50, -1.5*deg, 1.5*deg, 50, 0, 2*deg)
    simulation.setBeamParameters(0.1*nm, incident_angle, 0)
    simulation.beam().setIntensity(1e+08)
    simulation.setSample(get_sample(params))
    return simulation


def simulation1(params):
    params["incident_angle"] = 0.1*deg
    return get_simulation(params)


def simulation2(params):
    params["incident_angle"] = 0.4*deg
    return get_simulation(params)


def create_real_data(incident_alpha):
    """
    Generating "real" data by adding noise to the simulated data.
    """
    params = {
        'radius_a': 5*nm,
        'radius_b': 6*nm,
        'height': 8*nm,
        "incident_angle": incident_alpha
    }

    simulation = get_simulation(params)
    simulation.runSimulation()

    # retrieving simulated data in the form of numpy array
    real_data = simulation.result().array()

    # spoiling simulated data with the noise to produce "real" data
    noise_factor = 0.1
    noisy = np.random.normal(real_data, noise_factor*np.sqrt(real_data))
    noisy[noisy < 0.1] = 0.1
    return noisy


class PlotObserver():
    """
    Draws fit progress every nth iteration. Real data, simulated data
    and chi2 map will be shown for both datasets.
    """
    def __init__(self):
        self.fig = plt.figure(figsize=(12.8, 10.24))
        self.fig.canvas.draw()

    def __call__(self, fit_objective):
        self.update(fit_objective)

    @staticmethod
    def plot_dataset(fit_objective, canvas):
        for i_dataset in range(0, fit_objective.fitObjectCount()):
            real_data = fit_objective.experimentalData(i_dataset)
            simul_data = fit_objective.simulationResult(i_dataset)
            chi2_map = fit_objective.relativeDifference(i_dataset)

            zmax = real_data.histogram2d().getMaximum()

            plt.subplot(canvas[i_dataset*3])
            bp.plot_colormap(real_data,
                             title="\"Real\" data - #" +
                             str(i_dataset + 1),
                             intensity_min=1,
                             intensity_max=zmax,
                             zlabel="")
            plt.subplot(canvas[1 + i_dataset*3])
            bp.plot_colormap(simul_data,
                             title="Simulated data - #" +
                             str(i_dataset + 1),
                             intensity_min=1,
                             intensity_max=zmax,
                             zlabel="")
            plt.subplot(canvas[2 + i_dataset*3])
            bp.plot_colormap(chi2_map,
                             title="Chi2 map - #" + str(i_dataset + 1),
                             intensity_min=0.001,
                             intensity_max=10,
                             zlabel="")

    @staticmethod
    def display_fit_parameters(fit_objective):
        """
        Displays fit parameters, chi and iteration number.
        """
        plt.title('Parameters')
        plt.axis('off')

        iteration_info = fit_objective.iterationInfo()

        plt.text(
            0.01, 0.85, "Iterations  " +
            '{:d}'.format(iteration_info.iterationCount()))
        plt.text(0.01, 0.75,
                 "Chi2       " + '{:8.4f}'.format(iteration_info.chi2()))
        for index, params in enumerate(iteration_info.parameters()):
            plt.text(
                0.01, 0.55 - index*0.1,
                '{:30.30s}: {:6.3f}'.format(params.name(), params.value))

    @staticmethod
    def plot_fit_parameters(fit_objective):
        """
        Displays fit parameters, chi and iteration number.
        """
        plt.axis('off')

        iteration_info = fit_objective.iterationInfo()

        plt.text(
            0.01, 0.95, "Iterations  " +
            '{:d}'.format(iteration_info.iterationCount()))
        plt.text(0.01, 0.70,
                 "Chi2       " + '{:8.4f}'.format(iteration_info.chi2()))
        for index, params in enumerate(iteration_info.parameters()):
            plt.text(
                0.01, 0.30 - index*0.3,
                '{:30.30s}: {:6.3f}'.format(params.name(), params.value))

    def update(self, fit_objective):
        self.fig.clf()

        # we divide figure to have 3x3 subplots, with two first rows occupying
        # most of the space
        canvas = matplotlib.gridspec.GridSpec(3,
                                              3,
                                              width_ratios=[1, 1, 1],
                                              height_ratios=[4, 4, 1])
        canvas.update(left=0.05, right=0.95, hspace=0.5, wspace=0.2)

        self.plot_dataset(fit_objective, canvas)
        plt.subplot(canvas[6:])
        self.plot_fit_parameters(fit_objective)

        plt.draw()
        plt.pause(0.01)


def run_fitting():
    """
    main function to run fitting
    """

    data1 = create_real_data(0.1*deg)
    data2 = create_real_data(0.4*deg)

    fit_objective = ba.FitObjective()
    fit_objective.addSimulationAndData(simulation1, data1, 1)
    fit_objective.addSimulationAndData(simulation2, data2, 1)
    fit_objective.initPrint(10)

    # creating custom observer which will draw fit progress
    plotter = PlotObserver()
    fit_objective.initPlot(10, plotter.update)

    params = ba.Parameters()
    params.add("radius_a", 4.*nm, min=2, max=10)
    params.add("radius_b", 6.*nm, vary=False)
    params.add("height", 4.*nm, min=2, max=10)

    minimizer = ba.Minimizer()
    result = minimizer.minimize(fit_objective.evaluate, params)
    fit_objective.finalize(result)


if __name__ == '__main__':
    run_fitting()
    plt.show()



# Functional test: tests of IntensityData object

import numpy, os, sys, unittest
import bornagain as ba


class IntensityDataTest(unittest.TestCase):
    """
    Test creation of ba.IntensityData object and different ways of accessing its data
    """
    def test_empty_data(self):
        data = ba.IntensityData()
        self.assertEqual(1, data.getAllocatedSize())
        self.assertEqual(0, data.rank())
        self.assertEqual(0, data.totalSum())

    def test_create_1d_output_data_from_numpy(self):
        input = numpy.array([0, 1, 2, 3, 4, 5, 6], dtype=float)
        output_data = ba.importArrayToOutputData(input)
        output = output_data.getArray()
        self.assertEqual(input.size, output.size)
        numpy.testing.assert_allclose(input, output)

    def test_create_2d_output_data_from_numpy(self):
        input = numpy.array([[0, 1], [2, 3]], dtype=float)
        output_data = ba.importArrayToOutputData(input)
        output = output_data.getArray()
        self.assertEqual(input.size, output.size)
        numpy.testing.assert_allclose(input, output)

    def test_create_1d_object(self):
        axis0 = ba.FixedBinAxis("angle", 20, 0, 20.)
        self.assertEqual(20, axis0.size())
        self.assertEqual(0, axis0.lowerBound())
        self.assertEqual(20, axis0.upperBound())
        data = ba.IntensityData()
        data.addAxis(axis0)
        self.assertEqual(20, data.getAllocatedSize())
        self.assertEqual(1, data.rank())
        self.assertEqual(0, data.totalSum())

    def test_create_2d_object(self):
        data = ba.IntensityData()
        data.addAxis("axis0", 10, 0, 10)
        data.addAxis("axis1", 20, 0, 20)
        self.assertEqual(200, data.getAllocatedSize())
        self.assertEqual(2, data.rank())
        self.assertEqual(0, data.totalSum())
        data.setAllTo(1)
        self.assertEqual(200, data.totalSum())
        for i in range(0, data.getAllocatedSize()):
            data[i] = data[i]* -1.0
        self.assertEqual(-200, data.totalSum())

    def test_access_simulation_intensity(self):
        simulation = ba.GISASSimulation()
        simulation.setDetectorParameters(10, -1, 1, 100, 0, 2)
        data = simulation.result().histogram2d()
        self.assertEqual(1000, data.getTotalNumberOfBins())
        self.assertEqual(2, data.rank())
        self.assertEqual(0, data.integral())
        self.assertEqual(10, data.xAxis().size())

    def test_numpy_array(self):
        data = ba.IntensityData()
        data.addAxis("axis0", 10, 0, 10)
        data.addAxis("axis1", 20, 0, 20)
        data.setAllTo(1)
        self.assertEqual((20, 10), data.getArray().shape)
        self.assertEqual((data.totalSum()), numpy.sum(data.getArray()))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
Plots intensity data difference stored in BornAgain "*.int" or "*.int.gz" format
Usage: python plot_diff_int.py reference.int.gz new.int.gz
"""

import numpy as np
import bornagain as ba
import plot_int as pid
import sys


def plot_diff_int(filename1, filename2):
    intensity_ref = ba.IntensityDataIOFactory.readIntensityData(filename1)
    intensity_other = ba.IntensityDataIOFactory.readIntensityData(filename2)
    data = 2 * np.abs(intensity_ref.array() - intensity_other.array()) \
           / (np.abs(intensity_ref.array()) + np.abs(intensity_other.array()))
    if data.max() == 0:
        exit("Both data sets are equal, there is nothing to plot.")
    rank = intensity_ref.rank()
    if rank == 2:
        pid.plot_raw_data_2d(data, [
            intensity_ref.getXmin()/ba.deg,
            intensity_ref.getXmax()/ba.deg,
            intensity_ref.getYmin()/ba.deg,
            intensity_ref.getYmax()/ba.deg
        ], data.max())
    elif rank == 1:
        axis_values = np.asarray(intensity_ref.xAxis().binCenters())/ba.deg
        pid.plot_raw_data_1d(axis_values, data, log_y=False)
    else:
        exit("Error in plot_diff_int: wrong data rank")


if __name__ == '__main__':
    if len(sys.argv) != 3:
        exit("Usage: plot_diff_int.py reference.int.gz other.int.gz")

    plot_diff_int(sys.argv[1], sys.argv[2])



#!/usr/bin/env python3
"""
Spherical particles embedded in the middle of the layer on top of substrate.
"""
import bornagain as ba
from bornagain import deg, nm, kvector_t


def get_sample():
    """
    Returns a sample with spherical particles in a layer
    between vacuum and substrate.
    """

    # Define materials
    material_IntermLayer = ba.HomogeneousMaterial("IntermLayer", 3.45e-06,
                                                  5.24e-09)
    material_Particle = ba.HomogeneousMaterial("Particle", 0, 0)
    material_Substrate = ba.HomogeneousMaterial("Substrate", 7.43e-06,
                                                1.72e-07)
    material_Vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)

    # Define form factors
    ff = ba.FormFactorFullSphere(10.2*nm)

    # Define particles
    particle = ba.Particle(material_Particle, ff)
    particle_position = kvector_t(0, 0, -25.2*nm)
    particle.setPosition(particle_position)

    # Define particle layouts
    layout = ba.ParticleLayout()
    layout.addParticle(particle)
    layout.setTotalParticleSurfaceDensity(0.01)

    # Define layers
    layer_1 = ba.Layer(material_Vacuum)
    layer_2 = ba.Layer(material_IntermLayer, 30*nm)
    layer_2.addLayout(layout)
    layer_3 = ba.Layer(material_Substrate)

    # Define sample
    sample = ba.MultiLayer()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)
    sample.addLayer(layer_3)

    return sample


def get_simulation(sample):
    beam = ba.Beam(1, 0.15*nm, ba.Direction(0.15*deg, 0))
    detector = ba.SphericalDetector(200, 2*deg, 0, 1*deg)
    simulation = ba.GISASSimulation(beam, sample, detector)
    return simulation


if __name__ == '__main__':
    import ba_plot
    sample = get_sample()
    simulation = get_simulation(sample)
    ba_plot.run_and_plot(simulation)



"""
Fitting scalar and residual based objective functions
"""
import sys
import os
import unittest
import bornagain as ba
import numpy as np


class Rosenbrock:
    def __init__(self):
        self.m_expected_minimum = 0.0
        self.m_expected_params = [1, 1.0]
        pass

    def objective_function(self, params):
        """
        Returns the value of Rosenbrock function
        """
        x = params["x"].value
        y = params["y"].value
        tmp1 = y - x*x
        tmp2 = 1 - x
        return 100*tmp1*tmp1 + tmp2*tmp2


def decaying_sin(params, x):
    amp = params['amp'].value
    phaseshift = params['phase'].value
    freq = params['frequency'].value
    decay = params['decay'].value
    return amp*np.sin(x*freq + phaseshift)*np.exp(-x*x*decay)


class DecayingSin:
    def __init__(self):
        self.m_x = np.linspace(0, 10, 100)
        self.m_params = ba.Parameters()
        self.m_params.add(ba.Parameter('amp', 10))
        self.m_params.add(ba.Parameter('decay', 0.05))
        self.m_params.add(ba.Parameter('phase', 1))
        self.m_params.add(ba.Parameter('frequency', 4))
        self.m_eps_data = np.linspace(0, 10, 100)
        self.m_eps_data.fill(0.01)
        self.m_data = decaying_sin(self.m_params, self.m_x)

    def objective_function(self, params):
        model = decaying_sin(params, self.m_x)
        return (self.m_data - model)/self.m_eps_data


class StandaloneFitTest(unittest.TestCase):
    def test_RosenbrockFit(self):
        """
        Testing fit of rosenbrock function
        """
        params = ba.Parameters()
        params.add(
            ba.Parameter("x", -1.2, ba.AttLimits.limited(-5, 5), 0.01))
        params.add(ba.Parameter("y", 1, ba.AttLimits.limited(-5, 5), 0.01))

        model = Rosenbrock()
        minimizer = ba.Minimizer()
        result = minimizer.minimize(model.objective_function, params)

        print(result.toString())

        # check found parameter values
        np.testing.assert_almost_equal(result.parameters().values(),
                                       model.m_expected_params, 3)

        # check found minimum
        np.testing.assert_almost_equal(result.minValue(),
                                       model.m_expected_minimum, 3)

    def test_DecayingSinFit(self):
        params = ba.Parameters()
        params.add(ba.Parameter('amp', 1, ba.AttLimits.positive()))
        params.add(ba.Parameter('decay', 0.1, ba.AttLimits.positive()))
        params.add(ba.Parameter('phase', 0.1, ba.AttLimits.limited(0,
                                                                   3.1)))
        params.add(ba.Parameter('frequency', 1, ba.AttLimits.positive()))

        model = DecayingSin()
        minimizer = ba.Minimizer()
        result = minimizer.minimize(model.objective_function, params)

        print(result.toString())

        # check found parameter values
        np.testing.assert_almost_equal(result.parameters().values(),
                                       model.m_params.values(), 3)


if __name__ == '__main__':
    unittest.main()



#  **************************************************************************  #
"""
#   BornAgain: simulate and fit reflection and scattering
#
#   @file      Wrap/Python/ba_plot.py
#   @brief     Python extensions of the SWIG-generated Python module bornagain.
#
#   @homepage  http://apps.jcns.fz-juelich.de/BornAgain
#   @license   GNU General Public License v3 or higher (see COPYING)
#   @copyright Forschungszentrum Juelich GmbH 2016
#   @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
"""
#  **************************************************************************  #

import os
import bornagain as ba
try:  # workaround for build servers
    import numpy as np
    from matplotlib import pyplot as plt
    from matplotlib import gridspec, colors
except Exception as e:
    print("In ba_plot.py: {:s}".format(str(e)))

label_fontsize = 16

# default values from environment variables
CMAP = "CMAP" in os.environ and os.environ["CMAP"] or 'inferno'
NOSHOW = "NOSHOW" in os.environ


def get_axes_limits(result, units):
    """
    Returns axes range as expected by pyplot.imshow.
    :param result: SimulationResult object from a Simulation
    :param units: units to use
    :return: axes ranges as a flat list
    """
    axis_infos = result.axisInfo(units)
    limits = [[axis_infos[i].m_min, axis_infos[i].m_max]
              for i in range(len(axis_infos))]
    flat_limits = [v for sublist in limits for v in sublist]

    return flat_limits


def translate_axis_label(label):
    """
    Formats an axis label into a LaTeX representation
    :param label: text representation of the axis label
    :return: LaTeX representation
    """
    label_dict = {
        'X [nbins]': r'$X \; $(bins)',
        'X [mm]': r'$X \; $(mm)',
        'Y [nbins]': r'$Y \; $(bins)',
        'Y [mm]': r'$Y \; $(mm)',
        'phi_f [rad]': r'$\varphi_f \; $(rad)',
        'phi_f [deg]': r'$\varphi_f \; $(deg)',
        'alpha_i [rad]': r'$\alpha_i \; $(rad)',
        'alpha_i [deg]': r'$\alpha_i \; $(deg)',
        'alpha_f [rad]': r'$\alpha_f \; $(rad)',
        'alpha_f [deg]': r'$\alpha_f \; $(deg)',
        'Qx [1/nm]': r'$Q_x \; $(nm$^{-1}$)',
        'Qy [1/nm]': r'$Q_y \; $(nm$^{-1}$)',
        'Qz [1/nm]': r'$Q_z \; $(nm$^{-1}$)',
        'Q [1/nm]': r'$Q \; $(nm$^{-1}$)',
        'Position [nm]': r'$Position \; $(nm)'
    }
    if label in label_dict.keys():
        return label_dict[label]
    else:
        return label


def get_axes_labels(result, units):
    """
    Returns axes range as expected by pyplot.imshow.
    :param result: SimulationResult object from a Simulation
    :param units: units to use
    :return: axes ranges as a flat list
    """
    axis_infos = result.axisInfo(units)
    labels = [
        translate_axis_label(axis_infos[i].m_name)
        for i in range(len(axis_infos))
    ]

    return labels


def plot_array(array, axes_limits=None, **kwargs):
    """
    Plots numpy array as a colormap in log scale.
    """

    zmax = kwargs.pop('intensity_max', np.amax(array))
    zmin = kwargs.pop('intensity_min', 1e-6*zmax)

    if zmin == zmax == 0.0:
        norm = colors.Normalize(0, 1)
    else:
        norm = colors.LogNorm(zmin, zmax)

    xlabel = kwargs.pop('xlabel', None)
    ylabel = kwargs.pop('ylabel', None)
    zlabel = kwargs.pop('zlabel', "Intensity")
    title = kwargs.pop('title', None)
    aspect = kwargs.pop('aspect', 'equal')
    cmap = kwargs.pop('cmap', CMAP)
    withCBar = kwargs.pop('with_cb', True)

    im = plt.imshow(array,
                    cmap=cmap,
                    norm=norm,
                    aspect=aspect,
                    extent=axes_limits,
                    **kwargs)
    if withCBar:
        cb = plt.colorbar(im, pad=0.025)

    if xlabel:
        plt.xlabel(xlabel, fontsize=label_fontsize)
    if ylabel:
        plt.ylabel(ylabel, fontsize=label_fontsize)
    if zlabel:
        cb.set_label(zlabel, size=label_fontsize)

    if title:
        plt.title(title)

    return im


def plot_histogram(hist, **kwargs):
    """
    Plots intensity data as color map
    :param intensity: Histogram2D object obtained from GISASSimulation
    :param zmin: Min value on amplitude's color bar
    :param zmax: Max value on amplitude's color bar
    """

    if not 'xlabel' in kwargs:
        kwargs['xlabel'] = translate_axis_label(hist.xAxis().getName())
    if not 'ylabel' in kwargs:
        kwargs['ylabel'] = translate_axis_label(hist.yAxis().getName())

    axes_limits = [
        hist.getXmin(),
        hist.getXmax(),
        hist.getYmin(),
        hist.getYmax()
    ]

    title = kwargs.pop('title', None)

    plot_array(hist.array(),
               title=title,
               axes_limits=axes_limits,
               **kwargs)


def plot_colormap(result, **kwargs):
    """
    Plots intensity data as color map
    :param result: SimulationResult from GISAS/OffSpecularSimulation
    :param zmin: Min value on amplitude's color bar
    :param zmax: Max value on amplitude's color bar
    """

    units = kwargs.pop('units', ba.Axes.DEFAULT)
    axes_limits = get_axes_limits(result, units)
    axes_labels = get_axes_labels(result, units)

    if not 'xlabel' in kwargs:
        kwargs['xlabel'] = axes_labels[0]
    if not 'ylabel' in kwargs:
        kwargs['ylabel'] = axes_labels[1]

    return plot_array(result.array(), axes_limits=axes_limits, **kwargs)


def plot_specular_simulation_result(result, **kwargs):
    """
    Plots intensity data for specular simulation result
    :param result: SimulationResult from SpecularSimulation
    :param ymin: minimal y-axis value to show
    :param ymax: maximum y-axis value to show
    :param units: units on the x-axis
    """

    units = kwargs.pop('units', ba.Axes.DEFAULT)
    intensity = result.array(units)
    x_axis = result.axis(units)

    ymax = kwargs.pop('intensity_max', np.amax(np.amax(intensity)*2))
    ymin = kwargs.pop('intensity_min',
                      max(np.amin(intensity)*0.5, 1e-18*ymax))

    xlabel = kwargs.pop('xlabel', get_axes_labels(result, units)[0])
    ylabel = kwargs.pop('ylabel', "Intensity")
    title = kwargs.pop('title', None)

    plt.semilogy(x_axis, intensity, **kwargs)

    plt.ylim([ymin, ymax])

    if xlabel:
        plt.xlabel(xlabel, fontsize=label_fontsize)
    if ylabel:
        plt.ylabel(ylabel, fontsize=label_fontsize)

    if title:
        plt.title(title)


def plot_simulation_result(result, **kwargs):
    """
    Draws simulation result and (optionally) shows the plot.
    :param result_: SimulationResult object obtained from GISAS/OffSpecular/SpecularSimulation
    :param intensity_min: Min value on amplitude's axis or color bar
    :param intensity_max: Max value on amplitude's axis or color bar
    :param units: units for plot axes
    :param noshow: don't plot to interactive device
    """
    noshow = kwargs.pop('noshow', NOSHOW)

    if len(result.array().shape
           ) == 1:  # 1D data, specular simulation assumed
        plot_specular_simulation_result(result, **kwargs)
    else:
        plot_colormap(result, **kwargs)
    plt.tight_layout()
    if not (noshow):
        plt.show()
    else:
        print("plot_simulation_result: noshow")


def run_and_plot(simulation, **kwargs):
    simulation.runSimulation()
    plot_simulation_result(simulation.result(), **kwargs)



import numpy, os, sys, unittest
import bornagain as ba


class Histogram2DTest(unittest.TestCase):
    def test_constructFromNumpyInt(self):
        """
        Testing construction of 2D histogram from numpy array.
        Automatic conversion under Python3 is not working, that's why it is float64 and not int64
        """
        arr = numpy.array(
            [[1, 2, 3, 4, 5], [6, 7, 8, 9, 10], [11, 12, 13, 14, 15]],
            numpy.float64)
        hist = ba.Histogram2D(arr)

        self.assertEqual(hist.getNbinsX(), 5)
        self.assertEqual(hist.getXmin(), 0)
        self.assertEqual(hist.getXmax(), 5)

        self.assertEqual(hist.getNbinsY(), 3)
        self.assertEqual(hist.getYmin(), 0)
        self.assertEqual(hist.getYmax(), 3)

        self.assertEqual(hist.binContent(0, 0), 11)
        self.assertEqual(hist.binContent(0, 1), 6)
        self.assertEqual(hist.binContent(4, 2), 5)

        arr_from_hist = hist.getArray()

        for (x, y), element in numpy.ndenumerate(arr):
            self.assertEqual(element, arr_from_hist[x][y])

    def test_constructFromNumpyDouble(self):
        """
        Testing construction of 2D histogram from numpy array
        """
        arr = numpy.array([[1, 2, 3, 4, 5.0], [6, 7, 8, 9, 10.0],
                           [11, 12, 13, 14, 15.0]],
                          dtype=numpy.float64)
        hist = ba.Histogram2D(arr)

        self.assertEqual(hist.getNbinsX(), 5)
        self.assertEqual(hist.getXmin(), 0)
        self.assertEqual(hist.getXmax(), 5)

        self.assertEqual(hist.getNbinsY(), 3)
        self.assertEqual(hist.getYmin(), 0)
        self.assertEqual(hist.getYmax(), 3)

        self.assertEqual(hist.binContent(0, 0), 11)
        self.assertEqual(hist.binContent(0, 1), 6)
        self.assertEqual(hist.binContent(4, 2), 5)

        arr_from_hist = hist.getArray()

        for (x, y), element in numpy.ndenumerate(arr):
            self.assertEqual(element, arr_from_hist[x][y])

    def test_constructAndAddFromNumpyInt(self):
        """
        Adding to the histogram content from numpy array
        """
        arr = numpy.array(
            [[1, 2, 3, 4, 5], [6, 7, 8, 9, 10], [11, 12, 13, 14, 15]],
            dtype=numpy.float64)
        print(type(arr))
        hist = ba.Histogram2D(arr)
        # adding same content once again
        hist.addContent(arr)
        arr_from_hist = hist.getArray()

        for (x, y), element in numpy.ndenumerate(arr):
            self.assertEqual(element*2, arr_from_hist[x][y])

    def test_constructAndAddFromNumpyDouble(self):
        """
        Adding to the histogram content from numpy array
        """
        arr = numpy.array([[1, 2, 3, 4, 5.0], [6, 7, 8, 9, 10.0],
                           [11, 12, 13, 14, 15.0]],
                          dtype=numpy.float64)
        hist = ba.Histogram2D(arr)
        # adding same content once again
        hist.addContent(arr)
        arr_from_hist = hist.getArray()

        for (x, y), element in numpy.ndenumerate(arr):
            self.assertEqual(element*2, arr_from_hist[x][y])

    def create_histogram(self, arr):
        """
        Returns newly created object
        """
        return ba.IHistogram.createFrom(arr)

    def test_createFromInt(self):
        """
        Testing newly create object
        """
        arr = numpy.array(
            [[1, 2, 3, 4, 5], [6, 7, 8, 9, 10], [11, 12, 13, 14, 15]],
            dtype=numpy.float64)
        hist = self.create_histogram(arr)
        arr_from_hist = hist.getArray()

        for (x, y), element in numpy.ndenumerate(arr):
            self.assertEqual(element, arr_from_hist[x][y])

    def test_createFromDouble(self):
        """
        Testing newly create object
        """
        arr = numpy.array([[1, 2, 3, 4, 5.0], [6, 7, 8, 9, 10.0],
                           [11, 12, 13, 14, 15.0]],
                          dtype=numpy.float64)
        hist = self.create_histogram(arr)
        arr_from_hist = hist.getArray()

        for (x, y), element in numpy.ndenumerate(arr):
            self.assertEqual(element, arr_from_hist[x][y])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
Minimal working fit examples: finds radius of sphere in Born approximation.
"""
import bornagain as ba
from bornagain import deg, angstrom, nm


def get_simulation(params):
    """
    Returns GISAS simulation for given set of parameters.
    """
    radius = params["radius"]

    sphere = ba.Particle(ba.HomogeneousMaterial("Particle", 6e-4, 2e-8),
                         ba.FormFactorFullSphere(radius))

    layer = ba.Layer(ba.HomogeneousMaterial("Vacuum", 0, 0))
    layer.addLayout(ba.ParticleLayout(sphere))
    multi_layer = ba.MultiLayer()
    multi_layer.addLayer(layer)

    simulation = ba.GISASSimulation()
    simulation.setDetectorParameters(100, -1*deg, 1*deg, 100, 0, 2*deg)
    simulation.setBeamParameters(1*angstrom, 0.2*deg, 0)
    simulation.setSample(multi_layer)

    return simulation


def real_data():
    """
    Generating "experimental" data by running simulation with default parameters.
    """
    simulation = get_simulation({'radius': 5*nm})
    simulation.runSimulation()
    return simulation.result().array()


def run_fitting():
    """
    Setups and runs fit.
    """
    fit_objective = ba.FitObjective()
    fit_objective.addSimulationAndData(get_simulation, real_data())
    fit_objective.initPrint(10)

    params = ba.Parameters()
    params.add("radius", 4.*nm, min=0.01)

    minimizer = ba.Minimizer()
    result = minimizer.minimize(fit_objective.evaluate, params)
    fit_objective.finalize(result)


if __name__ == '__main__':
    run_fitting()



#!/usr/bin/env python3
"""
All form factors available in BornAgain in the Born Approximation
"""
import numpy
import bornagain as ba
import ba_plot
from bornagain import deg, angstrom
from matplotlib import pyplot as plt

phi_min, phi_max = -2, 2.0
alpha_min, alpha_max = 0, 2.0

formfactors = [
    ba.FormFactorAnisoPyramid(20, 16, 13, 60*deg),
    ba.FormFactorBox(20, 16, 13),
    ba.FormFactorCantellatedCube(15, 6),
    ba.FormFactorCone(10, 13, 60*deg),
    ba.FormFactorCone6(10, 13, 60*deg),
    ba.FormFactorCuboctahedron(20, 13, 0.7, 60*deg),
    ba.FormFactorCylinder(8, 16),
    ba.FormFactorDodecahedron(5),
    ba.FormFactorEllipsoidalCylinder(8, 13, 16),
    ba.FormFactorFullSphere(8),
    ba.FormFactorFullSpheroid(10, 13),
    ba.FormFactorHemiEllipsoid(10, 6, 8),
    ba.FormFactorIcosahedron(8),
    ba.FormFactorPrism3(10, 13),
    ba.FormFactorPrism6(5, 11),
    ba.FormFactorPyramid(18, 13, 60*deg),
    ba.FormFactorCosineRippleBox(27, 20, 14),
    ba.FormFactorSawtoothRippleBox(36, 25, 14, 3),
    ba.FormFactorTetrahedron(15, 6, 60*deg),
    ba.FormFactorTruncatedCube(15, 6),
    ba.FormFactorTruncatedSphere(5, 7, 0),
    ba.FormFactorTruncatedSpheroid(7.5, 9, 1.2, 0),
]


def get_sample(formfactor):
    """
    Returns a one-layer sample that contains particles with given form factor.
    """
    # defining materials
    m_vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)
    m_particle = ba.HomogeneousMaterial("Particle", 6e-4, 2e-8)

    # collection of particles
    particle = ba.Particle(m_particle, formfactor)
    particle_layout = ba.ParticleLayout()
    particle_layout.addParticle(particle)

    vacuum_layer = ba.Layer(m_vacuum)
    vacuum_layer.addLayout(particle_layout)

    multi_layer = ba.MultiLayer()
    multi_layer.addLayer(vacuum_layer)
    return multi_layer


def get_simulation(sample):
    """
    Returns GISAXS simulation with standard beam and detector.
    """
    simulation = ba.GISASSimulation()
    simulation.setDetectorParameters(100, phi_min*deg, phi_max*deg, 100,
                                     alpha_min*deg, alpha_max*deg)
    simulation.setBeamParameters(1*angstrom, 0.2*deg, 0)
    simulation.setSample(sample)
    return simulation


def simulate(ff):
    """
    Runs simulation for one form factor, and returns simulated intensity pattern
    """
    sample = get_sample(ff)
    simulation = get_simulation(sample)
    simulation.runSimulation()
    return simulation.result()


def simulate_and_plot():
    """
    Run simulation one by one for every form factor from the list and plot results
    on a single canvas
    """

    fig = plt.figure(figsize=(12.80, 10.24))

    for nplot, ff in enumerate(formfactors):
        name = ff.__class__.__name__
        name = name.replace("FormFactor", "")
        print("Generating intensity map in BA for '{0}'".format(name))

        result = simulate(ff)

        # showing the result
        plt.subplot(5, 5, nplot + 1)
        plt.subplots_adjust(wspace=0.3, hspace=0.3)

        ba_plot.plot_colormap(result, xlabel="", ylabel="", zlabel="")

        plt.tick_params(axis='both', which='major', labelsize=8)
        plt.tick_params(axis='both', which='minor', labelsize=6)
        plt.xticks(numpy.arange(phi_min, phi_max + 0.0001, 1))
        plt.text(-0.1,
                 2.15,
                 name,
                 horizontalalignment='center',
                 verticalalignment='center',
                 fontsize=9)
    plt.show()


if __name__ == '__main__':
    simulate_and_plot()



#!/usr/bin/env python3
"""
Fitting example: looking for background and scale factors.

Real data contains some "unknown" background and scale factor.
In the fit we are trying to find cylinder radius and height,
scale and background factors.
"""

import numpy as np
from matplotlib import pyplot as plt
import bornagain as ba
import ba_fitmonitor
from bornagain import deg, angstrom, nm


def get_sample(params):
    """
    Build the sample representing cylinders on top of substrate without interference.
    """
    radius = params["radius"]
    height = params["height"]

    m_vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)
    m_substrate = ba.HomogeneousMaterial("Substrate", 6e-6, 2e-8)
    m_particle = ba.HomogeneousMaterial("Particle", 6e-4, 2e-8)

    cylinder_ff = ba.FormFactorCylinder(radius, height)
    cylinder = ba.Particle(m_particle, cylinder_ff)

    layout = ba.ParticleLayout()
    layout.addParticle(cylinder)

    vacuum_layer = ba.Layer(m_vacuum)
    vacuum_layer.addLayout(layout)

    substrate_layer = ba.Layer(m_substrate, 0)
    multi_layer = ba.MultiLayer()
    multi_layer.addLayer(vacuum_layer)
    multi_layer.addLayer(substrate_layer)
    return multi_layer


def get_simulation(params):
    """
    Create and return GISAXS simulation with beam and detector defined
    """
    background = params["background"]
    scale = params["scale"]

    simulation = ba.GISASSimulation()
    simulation.setDetectorParameters(100, -1*deg, 1*deg, 100, 0, 2*deg)
    simulation.setBeamParameters(1*angstrom, 0.2*deg, 0)
    simulation.beam().setIntensity(1e12*scale)
    simulation.setBackground(ba.ConstantBackground(background))
    simulation.setSample(get_sample(params))

    return simulation


def create_real_data():
    """
    Generating "real" data by adding noise, background and scale
    to the simulated data.
    Cylinder radius is set to 5nm, cylinder height to 10nm.
    During the fit we will try to find cylinder height and radius and
    scale, background factors.
    """

    params = {
        'radius': 5*nm,
        'height': 10*nm,
        'scale': 2,
        'background': 1000
    }

    simulation = get_simulation(params)
    simulation.runSimulation()

    # retrieving simulated data in the form of numpy array
    return simulation.result().array()


def run_fitting():
    """
    main function to run fitting
    """
    real_data = create_real_data()

    fit_objective = ba.FitObjective()
    fit_objective.addSimulationAndData(get_simulation, real_data, 1)

    fit_objective.initPrint(10)
    observer = ba_fitmonitor.PlotterGISAS()
    fit_objective.initPlot(10, observer)

    params = ba.Parameters()
    params.add("radius", 5.*nm, vary=False)
    params.add("height", 9.*nm, min=8.*nm, max=12.*nm)
    params.add("scale", 1.5, min=1, max=3)
    params.add("background", 200, min=100, max=2000, step=100)

    minimizer = ba.Minimizer()
    result = minimizer.minimize(fit_objective.evaluate, params)
    fit_objective.finalize(result)


if __name__ == '__main__':
    run_fitting()
    plt.show()



#!/usr/bin/env python3
"""
Core shell nanoparticles
"""
import bornagain as ba
from bornagain import deg, nm


def get_sample():
    """
    Returns a sample with box-shaped core-shell particles on a substrate.
    """

    # Define materials
    material_Core = ba.HomogeneousMaterial("Core", 6e-05, 2e-08)
    material_Shell = ba.HomogeneousMaterial("Shell", 0.0001, 2e-08)
    material_Vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)

    # Define form factors
    ff_1 = ba.FormFactorBox(12*nm, 12*nm, 7*nm)
    ff_2 = ba.FormFactorBox(16*nm, 16*nm, 8*nm)

    # Define particles
    particle_1 = ba.Particle(material_Core, ff_1)
    particle_2 = ba.Particle(material_Shell, ff_2)

    # Define core shell particles
    particle_3 = ba.ParticleCoreShell(particle_2, particle_1)

    # Define interference functions
    iff = ba.InterferenceFunctionNone()

    # Define particle layouts
    layout = ba.ParticleLayout()
    layout.addParticle(particle_3)
    layout.setInterferenceFunction(iff)
    layout.setTotalParticleSurfaceDensity(0.01)

    # Define layers
    layer = ba.Layer(material_Vacuum)
    layer.addLayout(layout)

    # Define sample
    sample = ba.MultiLayer()
    sample.addLayer(layer)

    return sample


def get_simulation(sample):
    beam = ba.Beam(1, 0.1*nm, ba.Direction(0.2*deg, 0))
    detector = ba.SphericalDetector(200, 2*deg, 0, 1*deg)
    simulation = ba.GISASSimulation(beam, sample, detector)
    return simulation


if __name__ == '__main__':
    import ba_plot
    sample = get_sample()
    simulation = get_simulation(sample)
    ba_plot.run_and_plot(simulation)



"""
Parametric model of a GISAS simulation.
The idealized sample model consists of dilute cylinders on a substrate.
"""

import bornagain as ba
from bornagain import deg, nm

mat_vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)
mat_substrate = ba.HomogeneousMaterial("Substrate", 6e-6, 2e-8)
mat_particle = ba.HomogeneousMaterial("Particle", 6e-4, 2e-8)


def get_sample(params):
    cylinder_height = params["cylinder_height"]
    cylinder_radius = params["cylinder_radius"]

    ff = ba.FormFactorCylinder(cylinder_radius, cylinder_height)
    cylinder = ba.Particle(mat_particle, ff)
    layout = ba.ParticleLayout()
    layout.addParticle(cylinder)

    layer_1 = ba.Layer(mat_vacuum)
    layer_1.addLayout(layout)
    layer_2 = ba.Layer(mat_substrate)

    sample = ba.MultiLayer()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)
    return sample


def get_simulation(params):
    beam = ba.Beam(10**params['lg(intensity)'], 0.1*nm,
                   ba.Direction(0.2*deg, 0))
    det = ba.SphericalDetector(100, -1.5*deg, 1.5*deg, 100, 0, 3*deg)
    sample = get_sample(params)

    simulation = ba.GISASSimulation(beam, sample, det)
    simulation.setBackground(
        ba.ConstantBackground(10**params['lg(background)']))

    return simulation

def start_parameters_1():
    params = ba.Parameters()
    params.add("lg(intensity)", 5)
    params.add("lg(background)", 1)
    params.add("cylinder_height", 6.*nm, min=0.01)
    params.add("cylinder_radius", 6.*nm, min=0.01)
    return params



# Mixture of cylinders and prisms without interference

from mpi4py import MPI # this line has to be first
import numpy
import matplotlib
import pylab
from libBornAgainCore import *

comm = MPI.COMM_WORLD
world_size = comm.Get_size()
world_rank = comm.Get_rank()


def get_sample():
    """
    Build and return the sample representing cylinders and pyramids on top of
    substrate without interference.
    """
    # defining materials
    m_air = HomogeneousMaterial("Air", 0.0, 0.0)
    m_substrate = HomogeneousMaterial("Substrate", 6e-6, 2e-8)
    m_particle = HomogeneousMaterial("Particle", 6e-4, 2e-8)

    # collection of particles
    cylinder_ff = FormFactorCylinder(5*nanometer, 5*nanometer)
    cylinder = Particle(m_particle, cylinder_ff)
    prism_ff = FormFactorPrism3(10*nanometer, 5*nanometer)
    prism = Particle(m_particle, prism_ff)
    particle_layout = ParticleLayout()
    particle_layout.addParticle(cylinder, 0.5)
    particle_layout.addParticle(prism, 0.5)

    # air layer with particles and substrate form multi layer
    air_layer = Layer(m_air)
    air_layer.addLayout(particle_layout)
    substrate_layer = Layer(m_substrate)
    multi_layer = MultiLayer()
    multi_layer.addLayer(air_layer)
    multi_layer.addLayer(substrate_layer)
    return multi_layer


def get_simulation():
    """
    Create and return GISAS simulation with beam and detector defined
    """
    simulation = GISASSimulation()
    simulation.setDetectorParameters(100, -1.0*degree, 1.0*degree, 100, 0.0*degree, 2.0*degree)
    simulation.setBeamParameters(1.0*angstrom, 0.2*degree, 0.0*degree)
    sample = get_sample()
    simulation.setSample(sample)

    return simulation


def run_simulation():
    """
    Run simulation and plot results
    """
    simulation = get_simulation()

    if(world_size == 1):
        print("Not an MPI environment, run with 'mpirun -n 4 python ompi_sim_example.py'")
        exit(0)

    simulation.runMPISimulation()

    if(world_rank == 0):
        sumresult = simulation.result().array()
        print(sumresult)
        # pylab.imshow(sumresult + 1, norm=matplotlib.colors.LogNorm(), extent=[-1.0, 1.0, 0, 2.0])
        # pylab.show()


if __name__ == '__main__':
    run_simulation()



#!/usr/bin/env python3
"""
An example of taking into account beam angular divergence
and beam footprint correction in reflectometry calculations.
"""
import numpy as np
import bornagain as ba
from bornagain import angstrom, deg
from os import path
from matplotlib import pyplot as plt

# input parameters
wavelength = 1.54*angstrom
alpha_i_min = 0  # min incident angle, deg
alpha_i_max = 2*deg  # max incident angle, rad
beam_sample_ratio = 0.01  # beam-to-sample size ratio

# convolution parameters
d_ang = 0.01*deg  # spread width for incident angle
n_sig = 3  # number of sigmas to convolve over
n_points = 25  # number of points to convolve over

# substrate (Si)
si_sld_real = 2.0704e-06  # \AA^{-2}
# layer parameters
n_repetitions = 10
# Ni
ni_sld_real = 9.4245e-06  # \AA^{-2}
d_ni = 70*angstrom
# Ti
ti_sld_real = -1.9493e-06  # \AA^{-2}
d_ti = 30*angstrom


def get_sample():
    # defining materials
    m_vacuum = ba.MaterialBySLD("Vacuum", 0, 0)
    m_ni = ba.MaterialBySLD("Ni", ni_sld_real, 0)
    m_ti = ba.MaterialBySLD("Ti", ti_sld_real, 0)
    m_substrate = ba.MaterialBySLD("SiSubstrate", si_sld_real, 0)

    vacuum_layer = ba.Layer(m_vacuum)
    ni_layer = ba.Layer(m_ni, d_ni)
    ti_layer = ba.Layer(m_ti, d_ti)
    substrate_layer = ba.Layer(m_substrate)
    multi_layer = ba.MultiLayer()
    multi_layer.addLayer(vacuum_layer)
    for i in range(n_repetitions):
        multi_layer.addLayer(ti_layer)
        multi_layer.addLayer(ni_layer)
    multi_layer.addLayer(substrate_layer)
    return multi_layer


def create_real_data():
    """
    Loading data from genx_angular_divergence.dat
    """
    filepath = path.join(path.dirname(path.realpath(__file__)),
                         "genx_angular_divergence.dat.gz")
    ax_values, real_data = np.loadtxt(filepath,
                                      usecols=(0, 1),
                                      skiprows=3,
                                      unpack=True)

    # translating axis values from double incident angle # to incident angle
    ax_values *= 0.5

    return ax_values, real_data


def get_simulation(sample, scan_size=500):
    """
    Returns a specular simulation with beam and detector defined.
    """
    footprint = ba.FootprintSquare(beam_sample_ratio)
    alpha_distr = ba.RangedDistributionGaussian(n_points, n_sig)

    scan = ba.AngularSpecScan(wavelength, scan_size, alpha_i_min,
                              alpha_i_max)
    scan.setFootprintFactor(footprint)
    scan.setAbsoluteAngularResolution(alpha_distr, d_ang)

    simulation = ba.SpecularSimulation()
    simulation.setScan(scan)
    simulation.setSample(sample)

    return simulation


if __name__ == '__main__':
    import ba_plot
    sample = get_sample()
    simulation = get_simulation(sample)
    ba_plot.run_and_plot(simulation)
    genx_axis, genx_values = create_real_data()
    plt.semilogy(genx_axis, genx_values, 'ko', markevery=300)
    plt.legend(['BornAgain', 'GenX'], loc='upper right')
    plt.show()



"""
Contains machinery to convert BornAgain filename to the type of the content
"""

class FileTypes:
    """
    Enumerator class to hold file types
    """
    CORE, FTEST, UTEST, PYCODE, MACROS, GUI, PYAPI, THIRD, UNDEF, TOTAL = range(10)
    descr=["Core", "Functional Tests", "Unit Tests", "*.py", "macros", "GUI", "PythonAPI", "Third", "Undef"]
    @staticmethod
    def loc_for_type():
        result = []
        for i in range(FileTypes.TOTAL):
            result.append(0)
        return result


def filePython(x):
    if ".py" in x and not ".pypp." in x: return True
    return False


def fileCpp(x):
    if ".h" in x or ".cpp" in x: return True
    return False


def dirCore(x):
    # if "/Core/Algorithms" in x: return True
    # if "/Core/FormFactors" in x: return True
    # if "/Core/Samples" in x: return True
    # if "/Core/StandardSamples" in x: return True
    # if "/Core/Tools" in x: return True
    # if "/Core/Fitting" in x: return True
    # if "/Core/inc" in x: return True
    # if "/Core/src" in x: return True
    # if "/Core/InputOutput" in x: return True
    # if "/Core/Geometry" in x: return True
    # if "/Fit/Factory" in x: return True
    # if "/Fit/FitKernel" in x: return True
    # if "/Fit/StandardFits" in x: return True
    # if "/Core/Geometry" in x: return True
    if "/Core/" in x: return True
    if "/Fit/" in x: return True

    return False


def dirPyAPI(x):
    if "/Core/PythonAPI" in x: return True
    if "/Fit/PythonAPI" in x: return True
    return False


def dirFuncTest(x):
    # if "/App/" in x: return True
    if "/Tests/Functional" in x: return True
    if "/Tests/PerformanceTests" in x: return True
    if "/Tests/FunctionalTests/TestCore" in x: return True
    if "/Tests/FunctionalTests/TestFit" in x: return True
    if "/Tests/FunctionalTests/TestPyCore" in x: return True
    if "/Tests/FunctionalTests/TestPyFit" in x: return True
    if "/Core/TestMachinery" in x: return True
    return False


def dirGUI(x):
    if "/GUI/coregui" in x  and not "widgetbox" in x and not "qttools" in x: return True
    if "/GUI/main" in x: return True
    if "/AppGUI/coregui" in x: return True
    if "/BASuite" in x: return True
    return False


def dirThirdParty(x):
    if "/ThirdParty" in x: return True
    return False


def dirSkip(x):
    if "/pub/core" in x: return True
    return False


def dirUnitTests(x):
    if "/UnitTests/" in x: return True
    if "/Tests/UnitTests/TestCore/" in x: return True
    if "/Tests/UnitTests/TestFit/" in x: return True
    if "/Tests/UnitTests/" in x: return True
    return False


def filetype(x):
    """
    Returns type of file
    """
    result = FileTypes.UNDEF

    if dirSkip(x):
        return result

    if dirPyAPI(x):
        result = FileTypes.PYAPI

    elif dirThirdParty(x):
        result = FileTypes.THIRD

    elif (fileCpp(x) or filePython(x)) and dirFuncTest(x):
        result = FileTypes.FTEST

    elif fileCpp(x) and dirCore(x):
        result = FileTypes.CORE

    elif dirUnitTests(x):
        result = FileTypes.UTEST

    elif dirGUI(x):
        result = FileTypes.GUI

    return result




#!/usr/bin/env python3
"""
Simple example demonstrating how polarized SANS experiments can be
simulated with BornAgain.
"""

import bornagain as ba
from bornagain import angstrom, deg, nm, nm2, kvector_t

# Magnetization of the particle's core material (A/m)
magnetization_core = kvector_t(0, 0, 1e7)


def get_sample():
    """
    Returns a sample with a magnetic core-shell particle in a solvent.
    """

    # Define materials
    magnetic_field = kvector_t(0, 0, 1e7)
    material_Core = ba.HomogeneousMaterial("Core", 6e-06, 2e-08,
                                           magnetic_field)
    material_Shell = ba.HomogeneousMaterial("Shell", 1e-07, 2e-08)
    material_Solvent = ba.HomogeneousMaterial("Solvent", 5e-06, 0)

    # Define form factors
    ff_1 = ba.FormFactorFullSphere(10*nm)
    ff_2 = ba.FormFactorFullSphere(12*nm)

    # Define particles
    particle_1 = ba.Particle(material_Core, ff_1)
    particle_1_position = kvector_t(0, 0, 2*nm)
    particle_1.setPosition(particle_1_position)
    particle_2 = ba.Particle(material_Shell, ff_2)

    # Define core shell particles
    particle = ba.ParticleCoreShell(particle_2, particle_1)

    # Define particle layouts
    layout = ba.ParticleLayout()
    layout.addParticle(particle)
    layout.setTotalParticleSurfaceDensity(0.01)

    # Define layers
    layer = ba.Layer(material_Solvent)
    layer.addLayout(layout)

    # Define sample
    sample = ba.MultiLayer()
    sample.addLayer(layer)

    return sample


def get_simulation(sample):
    """
    Returns a polarized SANS simulation
    """
    # Beam from above (perpendicular to sample):
    beam = ba.Beam(1, 0.4*nm, ba.Direction(90*deg, 0))
    beam.setPolarization(kvector_t(0, 0, 1))

    # Detector opposite to source:
    detPos = 2000  # distance from sample center to detector in mm
    detWid = 500  # detector width in mm
    detPix = 200  # number of pixels per direction
    det = ba.RectangularDetector(detPix, detWid, detPix, detWid)
    det.setPerpendicularToDirectBeam(detPos, detWid/2, detWid/2)
    det.setAnalyzerProperties(kvector_t(0, 0, -1), 1, 0.5)

    return ba.GISASSimulation(beam, sample, det)


if __name__ == '__main__':
    import ba_plot
    sample = get_sample()
    simulation = get_simulation(sample)
    ba_plot.run_and_plot(simulation) # TODO: restore units=ba.Axes.QSPACE)



"""
Testing python specific API for FitObjective related classes.
"""
import sys
import os
import unittest
import numpy as np
import bornagain as ba


class SimulationBuilder:
    def __init__(self):
        self.m_ncalls = 0
        self.m_pars = None
        self.m_nrow = 3
        self.m_ncol = 4

    def size(self):
        return self.m_nrow*self.m_ncol

    def build_simulation(self, pars):
        self.m_ncalls += 1
        self.m_pars = dict(pars)

        ml = ba.MultiLayer()
        material = ba.HomogeneousMaterial("Shell", 0, 0)
        ml.addLayer(ba.Layer(material))
        ml.addLayer(ba.Layer(material))

        simulation = ba.GISASSimulation()
        simulation.setSample(ml)
        simulation.setDetectorParameters(self.m_ncol, 0, 1, self.m_nrow, 0,
                                         1)
        return simulation

    def create_data(self):
        return np.ones(shape=(self.m_nrow, self.m_ncol), dtype=np.float64)


class FitObserver:
    def __init__(self):
        self.m_ncalls = 0
        self.m_iterations = []

    def update(self, fit_objective):
        self.m_ncalls += 1
        self.m_iterations.append(
            fit_objective.iterationInfo().iterationCount())


class FitObjectiveAPITest(unittest.TestCase):
    def test_SimulationBuilderCallback(self):
        """
        Testing simulation construction using Python callback
        """
        pars = ba.Parameters()
        pars.add(ba.Parameter("par0", 0))
        pars.add(ba.Parameter("par1", 1))

        builder = SimulationBuilder()
        data = builder.create_data()

        # adding simulation callback and experimental data
        objective = ba.FitObjective()
        objective.addSimulationAndData(builder.build_simulation, data, 1)
        self.assertEqual(builder.m_ncalls, 0)
        self.assertEqual(objective.numberOfFitElements(), 0)

        # running objective function
        objective.evaluate(pars)
        self.assertEqual(builder.m_ncalls, 1)
        self.assertEqual(objective.numberOfFitElements(), builder.size())
        self.assertEqual(builder.m_pars["par0"], 0)
        self.assertEqual(builder.m_pars["par1"], 1)

        # checking arrays of experimental and simulated data
        expected_sim = []
        expected_data = []
        for i in range(0, builder.size()):
            expected_sim.append(0)
            expected_data.append(1)
        self.assertEqual(expected_sim, list(objective.simulation_array()))
        self.assertEqual(expected_data,
                         list(objective.experimental_array()))

    def test_FittingObserver(self):
        """
        Testing simulation construction using Python callback
        """
        pars = ba.Parameters()
        pars.add(ba.Parameter("par0", 0))
        pars.add(ba.Parameter("par1", 1))

        # adding simulation callback and experimental data
        builder = SimulationBuilder()
        data = builder.create_data()
        objective = ba.FitObjective()
        objective.addSimulationAndData(builder.build_simulation, data, 1)

        # adding observer
        observer = FitObserver()
        objective.initPlot(5, observer.update)

        # running objective function 11 times
        for i in range(0, 11):
            objective.evaluate(pars)

        self.assertEqual(observer.m_ncalls, 3)
        self.assertEqual(observer.m_iterations, [1, 6, 11])

    def test_IterationInfo(self):
        """
        Testing map of parameters obtained from IterationInfo
        """

        params = ba.Parameters()
        params.add("bbb", 1)
        params.add("aaa", 2)

        info = ba.IterationInfo()
        info.update(params, 3)
        par_map = info.parameterMap()

        expected_names = ["aaa", "bbb"]
        expected_values = [2, 1.0]
        names = []
        values = []
        for key in par_map:
            names.append(key)
            values.append(par_map[key])

        self.assertEqual(names, expected_names)
        self.assertEqual(values, expected_values)
        self.assertEqual(info.iterationCount(), 1)


if __name__ == '__main__':
    unittest.main()



import sys, unittest, ctypes
import inspect
import bornagain as ba

initial_width = 42
initial_length = 42


class SampleBuilder(ba.ISampleBuilder):
    """
    Test class with one registered parameter and onChange() counter
    """
    def __init__(self):
        super(SampleBuilder, self).__init__()

        self.m_onchange_counter = 0

        self.m_width = ctypes.c_double(initial_width)
        self.registerParameter("width", ctypes.addressof(self.m_width))

    def onChange(self):
        self.m_onchange_counter += 1


class BuilderPrototype(ba.ISampleBuilder):
    """
    Prototype demonstrating new way of parameter registration
    """
    def __init__(self):
        super(BuilderPrototype, self).__init__()

        self.m_onchange_counter = 0

        self.m_length = 42
        self.register_prototype("length", "self.m_length")

    def onChange(self):
        self.m_onchange_counter += 1
        # update all class members with values from ctypes wrappers
        for key, value in self.__dict__.items():
            if "wrapper_" in key:
                original_key = key.split("wrapper_", 1).pop()
                self.__dict__[original_key] = value.value

    def register_prototype(self, par_name, var_key):
        var_key = var_key.split("self.", 1).pop()
        if not var_key in self.__dict__:
            raise Exception("Can't find variable self." + var_key +
                            " in given class")
        # defining the name for new ctype variable and creating dynamic attribute
        wrapper_key = "wrapper_" + var_key
        setattr(self, wrapper_key,
                ctypes.c_double(self.__dict__.get(var_key)))
        # registering new attribute in BornAgain
        self.registerParameter(
            par_name, ctypes.addressof(getattr(self, wrapper_key)))


class SampleBuilderTest(unittest.TestCase):
    def test_registerParameters(self):
        """
        Checking parameter registration and setting the value.
        """
        builder = SampleBuilder()

        self.assertEqual(builder.m_width.value, initial_width)
        builder.setParameterValue("width", 43)
        self.assertEqual(builder.m_width.value, 43)

    def test_parameterString(self):
        """
        Checking that parameterString() function is correctly exposed to Python
        """
        par_str = SampleBuilder().parametersToString()
        self.assertEqual(par_str, "'/SampleBuilder/width':42\n\n")

    def test_onChange(self):
        """
        Checking that onChange is triggered once when parameter is changed
        """
        builder = SampleBuilder()
        self.assertEqual(builder.m_onchange_counter, 0)
        builder.setParameterValue("width", 43)
        self.assertEqual(builder.m_onchange_counter, 1)
        builder.setParameterValue("width", 44)
        self.assertEqual(builder.m_onchange_counter, 2)

        # setting same value, onChange shouldn't be triggered
        builder.setParameterValue("width", 44)
        self.assertEqual(builder.m_onchange_counter, 2)

    def test_registerPrototype(self):
        builder = BuilderPrototype()

        self.assertEqual(builder.m_onchange_counter, 0)
        self.assertEqual(builder.m_length, initial_length)

        builder.setParameterValue("length", 43)
        self.assertEqual(builder.m_length, 43)
        self.assertEqual(builder.m_onchange_counter, 1)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
Large cylinders in DWBA.

This example demonstrates that for large particles (~1000nm) the form factor
oscillates rapidly within one detector bin and analytical calculations
(performed for the bin center) give completely wrong intensity pattern.
In this case Monte-Carlo integration over detector bin should be used.
"""
import bornagain as ba
from bornagain import deg, angstrom, nm
import ba_plot
from matplotlib import pyplot as plt

default_cylinder_radius = 10*nm
default_cylinder_height = 20*nm


def get_sample(cylinder_radius, cylinder_height):
    # Define materials
    m_vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)
    m_substrate = ba.HomogeneousMaterial("Substrate", 6e-6, 2e-8)
    m_particle = ba.HomogeneousMaterial("Particle", 6e-4, 2e-8)

    # Define particle layout
    cylinder_ff = ba.FormFactorCylinder(cylinder_radius, cylinder_height)
    cylinder = ba.Particle(m_particle, cylinder_ff)
    particle_layout = ba.ParticleLayout()
    particle_layout.addParticle(cylinder)

    # Define layers
    vacuum_layer = ba.Layer(m_vacuum)
    vacuum_layer.addLayout(particle_layout)
    substrate_layer = ba.Layer(m_substrate)

    # Define sample
    multi_layer = ba.MultiLayer()
    multi_layer.addLayer(vacuum_layer)
    multi_layer.addLayer(substrate_layer)
    return multi_layer


def get_simulation(sample, integration_flag):
    """
    Returns a GISAXS simulation with defined beam and detector.
    If integration_flag=True, the simulation will integrate over detector bins.
    """
    beam = ba.Beam(1, 1*angstrom, ba.Direction(0.2*deg, 0))
    det = ba.SphericalDetector(200, -2*deg, 2*deg, 200, 0, 2*deg)
    simulation = ba.GISASSimulation(beam, sample, det)
    simulation.getOptions().setMonteCarloIntegration(integration_flag, 50)
    if not "__no_terminal__" in globals():
        simulation.setTerminalProgressMonitor()
    return simulation


def simulate_and_plot():
    """
    Run simulation and plot results 4 times: for small and large cylinders,
    with and without integration
    """

    fig = plt.figure(figsize=(12.80, 10.24))

    # conditions to define cylinders scale factor and integration flag
    conditions = [{
        'title': "Small cylinders, analytical calculations",
        'scale': 1,
        'integration': False,
        'zmin': 1e-5,
        'zmax': 1e2
    }, {
        'title': "Small cylinders, Monte-Carlo integration",
        'scale': 1,
        'integration': True,
        'zmin': 1e-5,
        'zmax': 1e2
    }, {
        'title': "Large cylinders, analytical calculations",
        'scale': 100,
        'integration': False,
        'zmin': 1e-5,
        'zmax': 1e10
    }, {
        'title': "Large cylinders, Monte-Carlo integration",
        'scale': 100,
        'integration': True,
        'zmin': 1e-5,
        'zmax': 1e10
    }]

    # run simulation 4 times and plot results
    for i_plot, condition in enumerate(conditions):
        scale = condition['scale']
        integration_flag = condition['integration']

        sample = get_sample(default_cylinder_radius*scale,
                            default_cylinder_height*scale)
        simulation = get_simulation(sample, integration_flag)
        simulation.runSimulation()
        result = simulation.result()

        # plotting results
        plt.subplot(2, 2, i_plot + 1)
        plt.subplots_adjust(wspace=0.3, hspace=0.3)

        zmin = condition['zmin']
        zmax = condition['zmax']
        ba_plot.plot_colormap(result,
                              intensity_min=zmin,
                              intensity_max=zmax)

        plt.text(0,
                 2.1,
                 conditions[i_plot]['title'],
                 horizontalalignment='center',
                 verticalalignment='center',
                 fontsize=12)
    plt.show()


if __name__ == '__main__':
    simulate_and_plot()



"""
Creates picture with number of lines of code.
Usage: python3 lines_of_code.py

The command should be executed in the directory where lines_of_code.py is located
(i.e. in <BornAgain>/dev-tools/analyze)
"""
import sys
if sys.version_info < (3, 0):
    exit("Requires python3, exiting ...")
import os
from baloc import HistoryCollector
from baloc import history_plot
import matplotlib.pyplot as plt


gitlog_filename = "gitlog.tmp"


def process_loc_number(targetfolder="../..", gitlog=gitlog_filename):
    prevfolder = os.getcwd()
    os.chdir(targetfolder)

    collector = HistoryCollector()
    collector.run()
    collector.save_report(os.path.join(prevfolder, gitlog))

    os.chdir(prevfolder)


def plot_loc_number(gitlog=gitlog_filename):
    history_plot(gitlog)
    plt.show()


if __name__ == '__main__':
    print('-' * 80)
    print("Generating picture for number of lines of code")
    print('-' * 80)
    print(" ")
    print("Possible options:")
    print("[0] - Generate {0} and picture.".format(gitlog_filename))
    print("[1] - Generate only {0}.".format(gitlog_filename))
    print("[2] - Generate picture using existing {0}.".format(gitlog_filename))
    print("[3] - Exit")

    var = int(input("Enter your choice [0]: ") or "0")

    if var == 0:
        process_loc_number()
        plot_loc_number()
    elif var == 1:
        process_loc_number()
    elif var == 2:
        plot_loc_number()
    else:
        exit("Good bye")





"""
Testing python specific API for Minimizer related classes.
"""
import sys
import os
import unittest
import bornagain as ba


class TestMinimizerHelper:
    def __init__(self):
        self.m_ncalls = 0
        self.m_pars = None

    def objective_function(self, pars):
        self.m_ncalls += 1
        self.m_pars = pars
        return 42.0


class MinimizerAPITest(unittest.TestCase):
    def test_ParameterAttribute(self):
        """
        Testing p.value attribute
        """
        par = ba.Parameter("par", 1)
        self.assertEqual(par.value, 1)
        par.value = 42.0
        self.assertEqual(par.value, 42)

    def test_ParametersSetIterator(self):
        """
        Testing of python iterator over defined fit parameters.
        """

        pars = ba.Parameters()
        self.assertEqual(pars.size(), 0)

        pars.add(ba.Parameter("par0", 1, ba.AttLimits.limitless()))
        pars.add(ba.Parameter("par1", 2, ba.AttLimits.limitless()))
        expected_names = ["par0", "par1"]
        for index, p in enumerate(pars):
            self.assertEqual(p.name(), expected_names[index])

    def test_ParametersAdd(self):
        """
        Testing Parameters::add method
        """

        params = ba.Parameters()
        params.add("par0", 0)
        params.add("par1", 1, min=1)
        params.add("par2", 2, max=2)
        params.add("par3", 3, min=1, max=2)
        params.add("par4", 4, vary=False)

        self.assertTrue(params["par0"].limits().isLimitless())
        self.assertTrue(params["par1"].limits().isLowerLimited())
        self.assertEqual(params["par1"].limits().lowerLimit(), 1)
        self.assertTrue(params["par2"].limits().isUpperLimited())
        self.assertEqual(params["par2"].limits().upperLimit(), 2)

        self.assertTrue(params["par3"].limits().isLimited())
        self.assertEqual(params["par3"].limits().lowerLimit(), 1)
        self.assertEqual(params["par3"].limits().upperLimit(), 2)

        self.assertTrue(params["par4"].limits().isFixed())

    def test_SimpleMinimizer(self):
        minimizer = ba.Minimizer()
        minimizer.setMinimizer("Test")

        pars = ba.Parameters()
        pars.add(ba.Parameter("par0", 0))
        pars.add(ba.Parameter("par1", 1))
        pars.add(ba.Parameter("par2", 2))

        helper = TestMinimizerHelper()
        result = minimizer.minimize(helper.objective_function, pars)

        # return value of objective function was propagated to MinimizerResult
        self.assertEqual(result.minValue(), 42)

        # objective function was called twice
        #(once by test minimizer, and second time during return type deduction)
        self.assertEqual(helper.m_ncalls, 2)

        # starting values of fit parameters were correctly send to objective func
        self.assertEqual(list(helper.m_pars.values()), [0, 1, 2.0])


if __name__ == '__main__':
    unittest.main()
